from eventsourcing.domain.model.decorators import attribute
from eventsourcing.domain.model.entity import AbstractEntityRepository, TimestampedVersionedEntity, \
    EntityWithHashchain


class Example(EntityWithHashchain, TimestampedVersionedEntity):
    """
    An example event sourced domain model entity.
    """
    class Event(EntityWithHashchain.Event, TimestampedVersionedEntity.Event):
        """Supertype for events of example entities."""

    class Created(Event, EntityWithHashchain.Created, TimestampedVersionedEntity.Created):
        """Published when an Example is created."""

    class AttributeChanged(Event, TimestampedVersionedEntity.AttributeChanged):
        """Published when an Example is created."""

    class Discarded(Event, TimestampedVersionedEntity.Discarded):
        """Published when an Example is discarded."""

    class Heartbeat(Event, TimestampedVersionedEntity.Event):
        """Published when a heartbeat in the entity occurs (see below)."""
        def mutate(self, obj):
            """Update obj with values from self."""
            assert isinstance(obj, Example), obj
            obj._count_heartbeats += 1

    def __init__(self, foo='', a='', b='', **kwargs):
        super(Example, self).__init__(**kwargs)
        self._foo = foo
        self._a = a
        self._b = b
        self._count_heartbeats = 0

    @attribute
    def foo(self):
        """An example attribute."""

    @attribute
    def a(self):
        """An example attribute."""

    @attribute
    def b(self):
        """Another example attribute."""

    def beat_heart(self, number_of_beats=1):
        self.__assert_not_discarded__()
        while number_of_beats > 0:
            self.__trigger_event__(self.Heartbeat)
            number_of_beats -= 1

    def count_heartbeats(self):
        return self._count_heartbeats


class AbstractExampleRepository(AbstractEntityRepository):
    pass


def create_new_example(foo='', a='', b=''):
    """
    Factory method for example entities.

    :rtype: Example
    """
    return Example.__create__(foo=foo, a=a, b=b)



from django.db import models


class IntegerSequencedRecord(models.Model):

    id = models.BigAutoField(primary_key=True)

    # Sequence ID (e.g. an entity or aggregate ID).
    sequence_id = models.UUIDField()

    # Position (index) of item in sequence.
    position = models.BigIntegerField()

    # Topic of the item (e.g. path to domain event class).
    topic = models.TextField()

    # State of the item (serialized dict, possibly encrypted).
    data = models.TextField()

    class Meta:
        unique_together = (("sequence_id", "position"),)
        db_table = 'integer_sequenced_items'


class TimestampSequencedRecord(models.Model):

    id = models.BigAutoField(primary_key=True)

    # Sequence ID (e.g. an entity or aggregate ID).
    sequence_id = models.UUIDField()

    # Position (timestamp) of item in sequence.
    position = models.DecimalField(max_digits=24, decimal_places=6)

    # Topic of the item (e.g. path to domain event class).
    topic = models.TextField()

    # State of the item (serialized dict, possibly encrypted).
    data = models.TextField()

    class Meta:
        unique_together = (("sequence_id", "position"),)
        db_table = 'timestamp_sequenced_items'
        indexes = [
            models.Index(fields=['position'], name='position_idx'),
        ]


class SnapshotRecord(models.Model):

    id = models.BigAutoField(primary_key=True)

    # Sequence ID (e.g. an entity or aggregate ID).
    sequence_id = models.UUIDField()

    # Position (index) of item in sequence.
    position = models.BigIntegerField()

    # Topic of the item (e.g. path to domain event class).
    topic = models.TextField()

    # State of the item (serialized dict, possibly encrypted).
    data = models.TextField()

    class Meta:
        unique_together = (("sequence_id", "position"),)
        db_table = 'snapshots'


class StoredEventRecord(models.Model):

    id = models.BigAutoField(primary_key=True)

    # Sequence ID (e.g. an entity or aggregate ID).
    originator_id = models.UUIDField()

    # Position (index) of item in sequence.
    originator_version = models.BigIntegerField()

    # Topic of the item (e.g. path to domain event class).
    event_type = models.TextField()

    # State of the item (serialized dict, possibly encrypted).
    state = models.TextField()

    class Meta:
        unique_together = (("originator_id", "originator_version"),)
        db_table = 'stored_events'



from eventsourcing.tests.example_application_tests.base import ExampleApplicationTestCase
from eventsourcing.tests.sequenced_item_tests.test_cassandra_record_manager import \
    WithCassandraRecordManagers


class TestExampleApplicationWithCassandra(WithCassandraRecordManagers, ExampleApplicationTestCase):
    pass



from unittest import TestCase, skipIf
from uuid import uuid4

import sqlite3

from decimal import Decimal

import six


# # This doesn't work with Python 2.7, which returns only 2 decimal
# # places. Django seems to work though, somehow, to get 6 places.
# # @skipIf(six.PY2, "Skipping sqlite3 decimal field test on Python 2.7")
# class Test(TestCase):
#     """
#     This test case demonstrates that sqlite3 truncates decimal places
#     before the value of a decimal field is given to the converter,
#     when a decimal converter is registered, otherwise it returns a
#     float with full precision. Seems a little odd to give the converter
#     less than is returns without a converter?
#     """
#     def test(self):
#
#         # Set up sqlite3 table with decimal column.
#         conn = sqlite3.connect(':memory:', detect_types=sqlite3.PARSE_DECLTYPES)
#         c = conn.cursor()
#         c.execute("""
# CREATE TABLE t (
#         id BINARY(16) NOT NULL,
#         position DECIMAL(24,8) NOT NULL,
#         PRIMARY KEY (id)
# );
# """)
#         # Insert a float with 8 decimal places.
#         f = 1234567891.12345678
#         sql = 'INSERT INTO t VALUES ("{}", "{}")'.format(uuid4().hex, f)
#         c.execute(sql)
#         conn.commit()
#
#         # Get the value and check it's the same.
#         c.execute("SELECT position from t")
#         r = c.fetchone()
#         position = r[0]
#         # self.assertEqual(f, position)
#
#         # Now, register a decimal converter with sqlite3.
#         sqlite3.register_converter("decimal", lambda s: Decimal(s.decode()))
#
#         # Then, the value which seems to be in the database loses precision!
#         c.execute("SELECT position from t")
#         r = c.fetchone()
#         position = r[0]
#         self.assertEqual(Decimal('1234567891.12346'), position)



from eventsourcing.application import command, process, simple, snapshotting

from eventsourcing.infrastructure.sqlalchemy.factory import SQLAlchemyInfrastructureFactory
from eventsourcing.infrastructure.sqlalchemy.records import EntitySnapshotRecord, StoredEventRecord


class SimpleApplication(simple.SimpleApplication):
    infrastructure_factory_class = SQLAlchemyInfrastructureFactory
    stored_event_record_class = StoredEventRecord
    snapshot_record_class = EntitySnapshotRecord
    is_constructed_with_session = True

    def __init__(self, uri=None, pool_size=5, session=None, **kwargs):
        self.uri = uri
        self.pool_size = pool_size
        self.session = session
        super(SimpleApplication, self).__init__(**kwargs)

    def setup_infrastructure(self, *args, **kwargs):
        super(SimpleApplication, self).setup_infrastructure(
            session=self.session, uri=self.uri, pool_size=self.pool_size,
            *args, **kwargs
        )
        if self.datastore and self.session is None:
            self.session = self.datastore.session


class ApplicationWithSnapshotting(snapshotting.ApplicationWithSnapshotting, SimpleApplication):
    pass


class ProcessApplication(process.ProcessApplication, SimpleApplication):
    pass


class ProcessApplicationWithSnapshotting(process.ProcessApplicationWithSnapshotting, SimpleApplication):
    pass


class CommandProcess(command.CommandProcess, SimpleApplication):
    pass



import uuid
from unittest.case import TestCase

from eventsourcing.application.policies import PersistencePolicy
from eventsourcing.domain.model.aggregate import AggregateRoot
from eventsourcing.domain.model.decorators import attribute
from eventsourcing.exceptions import EventHashError, HeadHashError
from eventsourcing.infrastructure.eventsourcedrepository import EventSourcedRepository
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.sequenceditemmapper import SequencedItemMapper
from eventsourcing.infrastructure.sqlalchemy.manager import SQLAlchemyRecordManager
from eventsourcing.infrastructure.sqlalchemy.records import IntegerSequencedNoIDRecord
from eventsourcing.tests.sequenced_item_tests.test_sqlalchemy_record_manager import \
    WithSQLAlchemyRecordManagers
from eventsourcing.utils.topic import get_topic


class TestAggregateRootEvent(TestCase):
    def test_validate_aggregate_events(self):
        event1 = AggregateRoot.Created(
            originator_version=0,
            originator_id='1',
            originator_topic=get_topic(AggregateRoot)
        )
        event1.__check_hash__()

        # Chain another event.
        event2 = AggregateRoot.AttributeChanged(
            originator_version=1,
            originator_id='1',
            __previous_hash__=event1.__event_hash__
        )
        event2.__check_hash__()

        # Chain another event.
        event3 = AggregateRoot.AttributeChanged(
            originator_version=2,
            originator_id='1',
            __previous_hash__=event2.__event_hash__
        )
        event3.__check_hash__()

    def test_event_hash_error(self):
        event1 = AggregateRoot.Created(
            originator_version=0,
            originator_id='1',
            originator_topic=get_topic(AggregateRoot)
        )
        event1.__check_hash__()

        # Break the hash.
        event1.__dict__['event_hash'] = 'damage'
        with self.assertRaises(EventHashError):
            event1.__check_hash__()


class TestExampleAggregateRoot(WithSQLAlchemyRecordManagers):
    def setUp(self):
        super(TestExampleAggregateRoot, self).setUp()
        self.app = ExampleDDDApplication(self.datastore)

    def tearDown(self):
        self.app.close()
        super(TestExampleAggregateRoot, self).tearDown()

    def test_aggregate1_lifecycle(self):
        # Create a new aggregate.
        aggregate = self.app.create_aggregate1()

        self.assertIsInstance(aggregate, Aggregate1)

        # Check it's got a head hash.
        self.assertTrue(aggregate.__head__)
        last_next_hash = aggregate.__head__

        # Check it does not exist in the repository.
        self.assertNotIn(aggregate.id, self.app.aggregate1_repository)

        # Save the aggregate.
        aggregate.__save__()

        # Check it now exists in the repository.
        self.assertIn(aggregate.id, self.app.aggregate1_repository)

        # Change an attribute of the aggregate root entity.
        self.assertNotEqual(aggregate.foo, 'bar')
        aggregate.foo = 'bar'
        self.assertEqual(aggregate.foo, 'bar')

        # Check the head hash has changed.
        self.assertNotEqual(aggregate.__head__, last_next_hash)
        last_next_hash = aggregate.__head__

        self.assertIn(aggregate.id, self.app.aggregate1_repository)

        self.assertNotEqual(self.app.aggregate1_repository[aggregate.id].foo, 'bar')
        aggregate.__save__()
        self.assertEqual(self.app.aggregate1_repository[aggregate.id].foo, 'bar')

        # Check the aggregate has zero entities.
        self.assertEqual(aggregate.count_examples(), 0)

        # Check the aggregate has zero entities.
        self.assertEqual(aggregate.count_examples(), 0)

        # Ask the aggregate to create an entity within itself.
        aggregate.create_new_example()

        # Check the aggregate has one entity.
        self.assertEqual(aggregate.count_examples(), 1)

        # Check the aggregate in the repo still has zero entities.
        self.assertEqual(self.app.aggregate1_repository[aggregate.id].count_examples(), 0)

        # Check the head hash has changed.
        self.assertNotEqual(aggregate.__head__, last_next_hash)
        last_next_hash = aggregate.__head__

        # Call save().
        aggregate.__save__()

        # Check the aggregate in the repo now has one entity.
        self.assertEqual(self.app.aggregate1_repository[aggregate.id].count_examples(), 1)

        # Create two more entities within the aggregate.
        aggregate.create_new_example()
        aggregate.create_new_example()

        # Save both "entity created" events in one atomic transaction.
        aggregate.__save__()

        # Check the aggregate in the repo now has three entities.
        self.assertEqual(self.app.aggregate1_repository[aggregate.id].count_examples(), 3)

        # Discard the aggregate, calls save().
        aggregate.__discard__()
        aggregate.__save__()

        # Check the next hash has changed.
        self.assertNotEqual(aggregate.__head__, last_next_hash)

        # Check the aggregate no longer exists in the repo.
        self.assertNotIn(aggregate.id, self.app.aggregate1_repository)

    def test_both_types(self):
        # Create a new aggregate.
        aggregate1 = self.app.create_aggregate1()
        aggregate2 = self.app.create_aggregate2()

        aggregate1.__save__()
        aggregate2.__save__()

        self.assertIsInstance(aggregate1, Aggregate1)
        self.assertIsInstance(aggregate2, Aggregate2)

        self.assertEqual(aggregate1.foo, '')
        self.assertEqual(aggregate2.foo, '')

        aggregate1.foo = 'bar'
        aggregate2.foo = 'baz'

        aggregate1.__save__()
        aggregate2.__save__()

        aggregate1 = self.app.aggregate1_repository[aggregate1.id]
        aggregate2 = self.app.aggregate2_repository[aggregate2.id]

        self.assertIsInstance(aggregate1, Aggregate1)
        self.assertIsInstance(aggregate2, Aggregate2)

        self.assertEqual(aggregate1.foo, 'bar')
        self.assertEqual(aggregate2.foo, 'baz')

        aggregate1.__discard__()
        aggregate1.__save__()
        self.assertFalse(aggregate1.id in self.app.aggregate1_repository)
        self.assertTrue(aggregate2.id in self.app.aggregate2_repository)

        # Todo: Somehow avoid all IDs existing in all repositories.
        # - either namespace the UUIDs, with a UUID for each type,
        #     with adjustments to repository and factory methods.
        # - or make sequence type be a thing, with IDs being valid within the type
        #     compound partition key in Cassandra,
        # self.assertFalse(aggregate2.id in self.app.aggregate1_repository)

    def test_validate_previous_hash_error(self):
        # Check event has valid originator head.
        aggregate = Aggregate1(id='1', foo='bar', __created_on__=0, __version__=0)
        event = Aggregate1.AttributeChanged(name='foo', value='bar', originator_id='1',
                                            originator_version=1, __previous_hash__=aggregate.__head__)
        event.__check_obj__(aggregate)

        # Check OriginatorHeadError is raised if the originator head is wrong.
        event.__dict__['__previous_hash__'] += 'damage'
        with self.assertRaises(HeadHashError):
            event.__check_obj__(aggregate)


class ExampleAggregateRoot(AggregateRoot):
    class Event(AggregateRoot.Event):
        """Supertype for events of example aggregates."""

    class Created(Event, AggregateRoot.Created):
        """Published when an ExampleAggregateRoot is created."""

    class AttributeChanged(Event, AggregateRoot.AttributeChanged):
        """Published when an ExampleAggregateRoot is changed."""

    class Discarded(Event, AggregateRoot.Discarded):
        """Published when an ExampleAggregateRoot is discarded."""

    class ExampleCreated(Event):
        """Published when an example entity is created within the aggregate."""

        def __init__(self, entity_id, **kwargs):
            super(ExampleAggregateRoot.ExampleCreated, self).__init__(entity_id=entity_id, **kwargs)

        @property
        def entity_id(self):
            return self.__dict__['entity_id']

        def __mutate__(self, aggregate):
            super(ExampleAggregateRoot.ExampleCreated, self).__mutate__(aggregate)
            entity = Example(entity_id=self.entity_id)
            aggregate._entities[entity.id] = entity
            return aggregate


class Aggregate1(ExampleAggregateRoot):
    def __init__(self, foo='', **kwargs):
        super(Aggregate1, self).__init__(**kwargs)
        self._entities = {}
        self._foo = foo

    @attribute
    def foo(self):
        """Simple event sourced attribute called 'foo'."""

    def create_new_example(self):
        assert not self.__is_discarded__
        self.__trigger_event__(self.ExampleCreated, entity_id=uuid.uuid4())

    def count_examples(self):
        return len(self._entities)


class Aggregate2(ExampleAggregateRoot):
    def __init__(self, foo='', **kwargs):
        super(Aggregate2, self).__init__(**kwargs)
        self._entities = {}
        self._foo = foo

    @attribute
    def foo(self):
        """Simple event sourced attribute called 'foo'."""

    def create_new_example(self):
        assert not self.__is_discarded__
        self.__trigger_event__(self.ExampleCreated, entity_id=uuid.uuid4())

    def count_examples(self):
        return len(self._entities)


class AggregateRepository(EventSourcedRepository):
    pass


class Example(object):
    """
    Example domain entity.
    """

    def __init__(self, entity_id):
        self._id = entity_id

    @property
    def id(self):
        return self._id


class ExampleDDDApplication(object):
    def __init__(self, datastore):
        event_store = EventStore(
            record_manager=SQLAlchemyRecordManager(
                session=datastore.session,
                record_class=IntegerSequencedNoIDRecord,
            ),
            sequenced_item_mapper=SequencedItemMapper(
                sequence_id_attr_name='originator_id',
                position_attr_name='originator_version',
            )
        )
        self.aggregate1_repository = AggregateRepository(
            event_store=event_store,
        )
        self.aggregate2_repository = AggregateRepository(
            event_store=event_store,
        )
        self.persistence_policy = PersistencePolicy(
            event_type=ExampleAggregateRoot.Event,
            event_store=event_store,
        )

    def create_aggregate1(self):
        """
        Factory method, creates and returns a new aggregate1 root entity.

        :rtype: Aggregate1
        """
        return Aggregate1.__create__()

    def create_aggregate2(self):
        """
        Factory method, creates and returns a new aggregate1 root entity.

        :rtype: Aggregate2
        """
        return Aggregate2.__create__()

    def close(self):
        self.persistence_policy.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



"""
The entity module provides base classes for domain entities.
"""
from abc import ABCMeta, abstractmethod
from uuid import uuid4

from six import with_metaclass

from eventsourcing.domain.model.events import AttributeChanged, Created, Discarded, DomainEvent, \
    EventWithOriginatorID, \
    EventWithOriginatorVersion, EventWithTimestamp, GENESIS_HASH, QualnameABC, publish, EventWithHash
from eventsourcing.exceptions import EntityIsDiscarded, HeadHashError, OriginatorIDError, \
    OriginatorVersionError
from eventsourcing.utils.times import decimaltimestamp_from_uuid
from eventsourcing.utils.topic import get_topic, resolve_topic


class DomainEntity(QualnameABC):
    """
    Base class for domain entities.
    """
    def __init__(self, id):
        self._id = id
        self.__is_discarded__ = False

    @property
    def id(self):
        """Entity ID allows an entity instance to be
        referenced and distinguished from others, even
        though its state may change over time.
        """
        return self._id

    class Event(EventWithOriginatorID, DomainEvent):
        """
        Supertype for events of domain entities.
        """
        def __mutate__(self, obj):
            # Call super method.
            return super(DomainEntity.Event, self).__mutate__(obj)

        def __check_obj__(self, obj):
            """
            Checks obj state before mutating.
            """
            # Check obj is not None.
            assert obj is not None, "'obj' is None"

            # Check ID matches originator ID.
            if obj.id != self.originator_id:
                raise OriginatorIDError(
                    "'{}' not equal to event originator ID '{}'"
                    "".format(obj.id, self.originator_id)
                )

    @classmethod
    def __create__(cls, originator_id=None, event_class=None, **kwargs):
        if originator_id is None:
            originator_id = uuid4()
        event = (event_class or cls.Created)(
            originator_id=originator_id,
            originator_topic=get_topic(cls),
            **kwargs
        )
        obj = event.__mutate__()
        obj.__publish__(event)
        return obj

    class Created(Event, Created):
        """
        Published when an entity is created.
        """

        def __init__(self, originator_topic, **kwargs):
            super(DomainEntity.Created, self).__init__(
                originator_topic=originator_topic,
                **kwargs
            )

        @property
        def originator_topic(self):
            return self.__dict__['originator_topic']

        def __mutate__(self, entity_class=None):
            if entity_class is None:
                entity_class = resolve_topic(self.originator_topic)
            obj = entity_class(**self.__entity_kwargs__)
            return obj

        @property
        def __entity_kwargs__(self):
            kwargs = self.__dict__.copy()
            kwargs['id'] = kwargs.pop('originator_id')
            kwargs.pop('originator_topic', None)
            kwargs.pop('__event_topic__', None)
            return kwargs

    def __change_attribute__(self, name, value):
        """
        Changes named attribute with the given value,
        by triggering an AttributeChanged event.
        """
        self.__trigger_event__(self.AttributeChanged, name=name, value=value)

    class AttributeChanged(Event, AttributeChanged):
        """
        Published when a DomainEntity is discarded.
        """

        def __mutate__(self, obj):
            obj = super(DomainEntity.AttributeChanged, self).__mutate__(obj)
            setattr(obj, self.name, self.value)
            return obj

    def __discard__(self):
        """
        Discards self, by triggering a Discarded event.
        """
        self.__trigger_event__(self.Discarded)

    class Discarded(Discarded, Event):
        """
        Published when a DomainEntity is discarded.
        """

        def __mutate__(self, obj):
            obj = super(DomainEntity.Discarded, self).__mutate__(obj)
            obj.__is_discarded__ = True
            return None

    def __assert_not_discarded__(self):
        """
        Raises exception if entity has been discarded already.
        """
        if self.__is_discarded__:
            raise EntityIsDiscarded("Entity is discarded")

    def __trigger_event__(self, event_class, **kwargs):
        """
        Constructs, applies, and publishes a domain event.
        """
        self.__assert_not_discarded__()
        event = event_class(
            originator_id=self._id,
            **kwargs
        )
        event.__mutate__(self)
        self.__publish__(event)

    def __publish__(self, event):
        """
        Publishes given event for subscribers in the application.

        :param event: domain event or list of events
        """
        self.__publish_to_subscribers__(event)

    def __publish_to_subscribers__(self, event):
        """
        Actually dispatches given event to publish-subscribe mechanism.

        :param event: domain event or list of events
        """
        publish(event)

    __hash__ = None  # For Python 2.7, so hash(obj) raises TypeError.

    def __eq__(self, other):
        return type(self) == type(other) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self.__eq__(other)


class EntityWithHashchain(DomainEntity):
    __genesis_hash__ = GENESIS_HASH

    def __init__(self, *args, **kwargs):
        super(EntityWithHashchain, self).__init__(*args, **kwargs)
        self.__head__ = type(self).__genesis_hash__

    class Event(EventWithHash, DomainEntity.Event):
        """
        Supertype for events of domain entities.
        """
        def __mutate__(self, obj):
            # Set entity head from event hash.
            obj.__head__ = self.__event_hash__

            # Call super method.
            return super(EntityWithHashchain.Event, self).__mutate__(obj)

        def __check_obj__(self, obj):
            """
            Extends superclass method by checking the __previous_hash__
            of this event matches the __head__ hash of the entity obj.
            """
            # Call super method.
            super(EntityWithHashchain.Event, self).__check_obj__(obj)

            # Check __head__ matches previous hash.
            if obj.__head__ != self.__dict__.get('__previous_hash__'):
                raise HeadHashError(obj.id, obj.__head__, type(self))

    class Created(Event, DomainEntity.Created):
        @property
        def __entity_kwargs__(self):
            # Call super method.
            kwargs = super(EntityWithHashchain.Created, self).__entity_kwargs__

            # Drop the event hashes.
            kwargs.pop('__event_hash__', None)
            kwargs.pop('__previous_hash__', None)

            return kwargs

        def __mutate__(self, entity_class=None):
            # Call super method.
            obj = super(EntityWithHashchain.Event, self).__mutate__(entity_class)

            # Set entity head from event hash.
            obj.__head__ = self.__event_hash__

            return obj

    @classmethod
    def __create__(cls, *args, **kwargs):
        kwargs['__previous_hash__'] = getattr(cls, '__genesis_hash__', GENESIS_HASH)
        return super(EntityWithHashchain, cls).__create__(*args, **kwargs)

    def __trigger_event__(self, event_class, **kwargs):
        assert isinstance(event_class, type), type(event_class)
        kwargs['__previous_hash__'] = self.__head__
        super(EntityWithHashchain, self).__trigger_event__(event_class, **kwargs)


class VersionedEntity(DomainEntity):
    def __init__(self, __version__=None, **kwargs):
        super(VersionedEntity, self).__init__(**kwargs)
        self.___version__ = __version__

    @property
    def __version__(self):
        return self.___version__

    def __trigger_event__(self, event_class, **kwargs):
        """
        Triggers domain event with entity's next version number.

        The event carries the version number that the originator
        will have when the originator is mutated with this event.
        (The event's originator version isn't the version of the
        originator that triggered the event. The Created event has
        version 0, and so a newly created instance is at version 0.
        The second event has originator version 1, and so will the
        originator when the second event has been applied.)
        """
        return super(VersionedEntity, self).__trigger_event__(
            event_class=event_class,
            originator_version=self.__version__ + 1,
            **kwargs
        )

    class Event(EventWithOriginatorVersion, DomainEntity.Event):
        """Supertype for events of versioned entities."""

        def __mutate__(self, obj):
            obj = super(VersionedEntity.Event, self).__mutate__(obj)
            if obj is not None:
                obj.___version__ = self.originator_version
            return obj

        def __check_obj__(self, obj):
            """
            Extends superclass method by checking the event's
            originator version follows (1 +) this entity's version.
            """
            super(VersionedEntity.Event, self).__check_obj__(obj)
            if self.originator_version != obj.__version__ + 1:
                raise OriginatorVersionError(
                    ("Event takes entity to version {}, "
                     "but entity is currently at version {}. "
                     "Event type: '{}', entity type: '{}', entity ID: '{}'"
                     "".format(self.originator_version, obj.__version__,
                               type(self).__name__, type(obj).__name__, obj._id)
                    )
                )

    class Created(DomainEntity.Created, Event):
        """Published when a VersionedEntity is created."""

        def __init__(self, originator_version=0, **kwargs):
            super(VersionedEntity.Created, self).__init__(originator_version=originator_version, **kwargs)

        @property
        def __entity_kwargs__(self):
            kwargs = super(VersionedEntity.Created, self).__entity_kwargs__
            kwargs['__version__'] = kwargs.pop('originator_version')
            return kwargs

    class AttributeChanged(Event, DomainEntity.AttributeChanged):
        """Published when a VersionedEntity is changed."""

    class Discarded(Event, DomainEntity.Discarded):
        """Published when a VersionedEntity is discarded."""

    # def __change_attribute__(self, name, value):
    #     if self.__version__:
    #         super(VersionedEntity, self).__change_attribute__(name, value)
    #     else:
    #         setattr(self, name, value)


class TimestampedEntity(DomainEntity):
    def __init__(self, __created_on__=None, **kwargs):
        super(TimestampedEntity, self).__init__(**kwargs)
        self.___created_on__ = __created_on__
        self.___last_modified__ = __created_on__

    @property
    def __created_on__(self):
        return self.___created_on__

    @property
    def __last_modified__(self):
        return self.___last_modified__

    class Event(DomainEntity.Event, EventWithTimestamp):
        """Supertype for events of timestamped entities."""

        def __mutate__(self, obj):
            """Update obj with values from self."""
            obj = super(TimestampedEntity.Event, self).__mutate__(obj)
            if obj is not None:
                assert isinstance(obj, TimestampedEntity), obj
                obj.___last_modified__ = self.timestamp
            return obj

    class Created(DomainEntity.Created, Event):
        """Published when a TimestampedEntity is created."""

        @property
        def __entity_kwargs__(self):
            kwargs = super(TimestampedEntity.Created, self).__entity_kwargs__
            kwargs['__created_on__'] = kwargs.pop('timestamp')
            return kwargs

    class AttributeChanged(Event, DomainEntity.AttributeChanged):
        """Published when a TimestampedEntity is changed."""

    class Discarded(Event, DomainEntity.Discarded):
        """Published when a TimestampedEntity is discarded."""


class TimeuuidedEntity(DomainEntity):
    def __init__(self, event_id, **kwargs):
        super(TimeuuidedEntity, self).__init__(**kwargs)
        self.___initial_event_id__ = event_id
        self.___last_event_id__ = event_id

    @property
    def __created_on__(self):
        return decimaltimestamp_from_uuid(self.___initial_event_id__)

    @property
    def __last_modified__(self):
        return decimaltimestamp_from_uuid(self.___last_event_id__)


class TimestampedVersionedEntity(TimestampedEntity, VersionedEntity):
    class Event(TimestampedEntity.Event, VersionedEntity.Event):
        """Supertype for events of timestamped, versioned entities."""

    class Created(TimestampedEntity.Created, VersionedEntity.Created, Event):
        """Published when a TimestampedVersionedEntity is created."""

    class AttributeChanged(Event, TimestampedEntity.AttributeChanged, VersionedEntity.AttributeChanged):
        """Published when a TimestampedVersionedEntity is created."""

    class Discarded(Event, TimestampedEntity.Discarded, VersionedEntity.Discarded):
        """Published when a TimestampedVersionedEntity is discarded."""


class TimeuuidedVersionedEntity(TimeuuidedEntity, VersionedEntity):
    pass


class AbstractEventPlayer(with_metaclass(ABCMeta)):
    pass


class AbstractEntityRepository(AbstractEventPlayer):
    @abstractmethod
    def __getitem__(self, entity_id):
        """
        Returns entity for given ID.
        """

    @abstractmethod
    def __contains__(self, entity_id):
        """
        Returns True or False, according to whether or not entity exists.
        """

    @abstractmethod
    def get_entity(self, entity_id, at=None):
        """
        Returns entity for given ID.
        """

    @property
    @abstractmethod
    def event_store(self):
        """
        Returns event store object used by this repository.
        """

    @abstractmethod
    def take_snapshot(self, entity_id, lt=None, lte=None):
        """
        Takes snapshot of entity state, using stored events.
        :return: Snapshot
        """



from eventsourcing.infrastructure.django.manager import DjangoRecordManager
from eventsourcing.infrastructure.django.models import IntegerSequencedRecord, TimestampSequencedRecord, SnapshotRecord
from eventsourcing.infrastructure.factory import InfrastructureFactory


class DjangoInfrastructureFactory(InfrastructureFactory):
    record_manager_class = DjangoRecordManager
    integer_sequenced_record_class = IntegerSequencedRecord
    timestamp_sequenced_record_class = TimestampSequencedRecord
    snapshot_record_class = SnapshotRecord



import unittest
from uuid import uuid4

from eventsourcing.application.policies import PersistencePolicy, SnapshottingPolicy
from eventsourcing.domain.model.entity import VersionedEntity, TimestampedEntity, AbstractEntityRepository
from eventsourcing.domain.model.events import publish
from eventsourcing.infrastructure.eventsourcedrepository import EventSourcedRepository
from eventsourcing.infrastructure.eventstore import AbstractEventStore

try:
    from unittest import mock
except:
    import mock


class TestPersistencePolicy(unittest.TestCase):
    def setUp(self):
        self.event_store = mock.Mock(spec=AbstractEventStore)
        self.persistence_policy = PersistencePolicy(
            event_store=self.event_store,
            event_type=VersionedEntity.Event
        )

    def tearDown(self):
        self.persistence_policy.close()

    def test_published_events_are_appended_to_event_store(self):
        # Check the event store's append method has NOT been called.
        assert isinstance(self.event_store, AbstractEventStore)
        self.assertEqual(0, self.event_store.append.call_count)

        # Publish a versioned entity event.
        entity_id = uuid4()
        domain_event1 = VersionedEntity.Event(
            originator_id=entity_id,
            originator_version=0,
        )
        publish(domain_event1)

        # Check the append method has been called once with the domain event.
        self.event_store.append.assert_called_once_with(domain_event1)

        # Publish a timestamped entity event (should be ignored).
        domain_event2 = TimestampedEntity.Event(
            originator_id=entity_id,
        )
        publish(domain_event2)

        # Check the append() has still only been called once with the first domain event.
        self.event_store.append.assert_called_once_with(domain_event1)


class TestSnapshottingPolicy(unittest.TestCase):
    def setUp(self):
        self.repository = mock.Mock(spec=AbstractEntityRepository)
        self.snapshot_store = mock.Mock(spec=AbstractEventStore)
        self.policy = SnapshottingPolicy(
            repository=self.repository,
            snapshot_store=self.snapshot_store,
            period=2,
        )

    def tearDown(self):
        self.policy.close()

    def test_published_events_are_appended_to_event_store(self):
        # Check the event store's append method has NOT been called.
        assert isinstance(self.repository, AbstractEntityRepository)
        self.assertEqual(0, self.repository.take_snapshot.call_count)

        # Publish a versioned entity event.
        entity_id = uuid4()
        domain_event1 = VersionedEntity.Event(
            originator_id=entity_id,
            originator_version=0,
        )
        domain_event2 = VersionedEntity.Event(
            originator_id=entity_id,
            originator_version=1,
        )

        # Check take_snapshot is called once for each event.
        publish(domain_event1)
        self.assertEqual(0, self.repository.take_snapshot.call_count)
        publish(domain_event2)
        self.assertEqual(1, self.repository.take_snapshot.call_count)

        # Check take_snapshot is called once for each list.
        publish([domain_event1, domain_event2])
        self.assertEqual(2, self.repository.take_snapshot.call_count)




import logging
import os
import time
import unittest

from eventsourcing.application.actors import Actors, shutdown_actor_system, start_actor_system, \
    start_multiproc_tcp_base_system
from eventsourcing.application.system import System
from eventsourcing.tests.test_system_fixtures import set_db_uri, create_new_order, Orders, Reservations, Payments

logger = logging.getLogger('')
logger.setLevel(logging.ERROR)
ch = logging.StreamHandler()
ch.setLevel(logging.ERROR)
logger.addHandler(ch)


class TestActors(unittest.TestCase):

    def test_simple_system_base(self):
        start_actor_system()
        self.check_actors()

    def test_multiproc_tcp_base(self):
        start_multiproc_tcp_base_system()
        self.check_actors()

    # def _test_multiproc_udp_base(self):
    #     start_multiproc_udp_base_system()
    #     self.check_actors(1, 1)
    #
    # def _test_multiproc_queue_base(self):
    #     start_multiproc_queue_base_system()
    #     self.check_actors()

    def check_actors(self, num_pipelines=2, num_orders_per_pipeline=5):

        pipeline_ids = list(range(num_pipelines))

        actors = Actors(self.system, pipeline_ids=pipeline_ids, shutdown_on_close=True)

        # Todo: Use wakeupAfter() to poll for new notifications (see Timer Messages).
        # Todo: Fix multiple pipelines with multiproc bases.

        order_ids = []

        with Orders(setup_table=True) as app, actors:

            # Create some new orders.
            for _ in range(num_orders_per_pipeline):

                for pipeline_id in pipeline_ids:
                    app.change_pipeline(pipeline_id)

                    order_id = create_new_order()
                    order_ids.append(order_id)

            # Wait for orders to be reserved and paid.
            retries = 100 + 100 * num_orders_per_pipeline * len(pipeline_ids)
            for i, order_id in enumerate(order_ids):

                while not app.repository[order_id].is_reserved:
                    time.sleep(0.1)
                    retries -= 1
                    assert retries, "Failed set order.is_reserved {} ({})".format(order_id, i)

                while retries and not app.repository[order_id].is_paid:
                    time.sleep(0.1)
                    retries -= 1
                    assert retries, "Failed set order.is_paid ({})".format(i)

            # Calculate timings from event timestamps.
            orders = [app.repository[oid] for oid in order_ids]
            first_timestamp = min([o.__created_on__ for o in orders])
            last_timestamp = max([o.__last_modified__ for o in orders])
            duration = last_timestamp - first_timestamp
            rate = len(order_ids) / float(duration)
            period = 1 / rate
            print("Orders system processed {} orders in {:.3f}s at rate of {:.1f} "
                  "orders/s, {:.3f}s each".format(len(order_ids), duration, rate, period))

            # Print min, average, max duration.
            durations = [o.__last_modified__ - o.__created_on__ for o in orders]
            print("Min order processing time: {:.3f}s".format(min(durations)))
            print("Mean order processing time: {:.3f}s".format(sum(durations) / len(durations)))
            print("Max order processing time: {:.3f}s".format(max(durations)))

    def setUp(self):
        # Shutdown base actor system, if running.
        # ActorSystem().shutdown()
        # Set environment.
        set_db_uri()
        # Define system.
        self.system = System(Orders | Reservations | Orders | Payments | Orders)

    def tearDown(self):
        # Unset environment.
        try:
            del (os.environ['DB_URI'])
        except KeyError:
            pass

        # Shutdown base actor system.
        shutdown_actor_system()



import os
from flask import Flask
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy_utils.types.uuid import UUIDType

from eventsourcing.example.application import get_example_application, init_example_application
from eventsourcing.infrastructure.sqlalchemy.manager import SQLAlchemyRecordManager

# Read DB URI from environment.
uri = os.environ.get('DB_URI', 'sqlite:///:memory:')


# Construct Flask application.
application = Flask(__name__)
application.config['SQLALCHEMY_DATABASE_URI'] = uri
application.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False

# Define database connection.
db = SQLAlchemy(application)


# Define database tables.
class IntegerSequencedItem(db.Model):
    __tablename__ = 'integer_sequenced_items'
    id = db.Column(db.BigInteger().with_variant(db.Integer, "sqlite"), primary_key=True)
    sequence_id = db.Column(UUIDType(), nullable=False)
    position = db.Column(db.BigInteger(), nullable=False)
    topic = db.Column(db.String(255), nullable=False)
    data = db.Column(db.Text(), nullable=False)
    __table_args__ = db.Index('index', 'sequence_id', 'position', unique=True),


# Construct eventsourcing application.
@application.before_first_request
def init_example_application_with_sqlalchemy():
    init_example_application(
        entity_record_manager=SQLAlchemyRecordManager(
            record_class=IntegerSequencedItem,
            session=db.session,
        )
    )


# Define Web application.
@application.route("/")
def hello():
    app = get_example_application()
    entity_id = app.create_new_example(foo='Hello There!').id
    entity = app.example_repository[entity_id]
    return "<h1 style='color:blue'>{}</h1>".format(entity.foo)


# Run directly, with uWSGI, or otherwise as a WSGI application.
#
# uwsgi -H PATH_TO_VIRTUALENV --master --processes 4 --threads 2 --wsgi-file PATH_TO_THIS_FILE --http :5001
#
if __name__ == "__main__":
    # Create tables.
    db.create_all()

    # Run the application.
    application.run(host='0.0.0.0', port=5001)



from eventsourcing.infrastructure.base import AbstractSequencedItemRecordManager
from eventsourcing.infrastructure.sequenceditem import SequencedItem


class InfrastructureFactory(object):
    record_manager_class = AbstractSequencedItemRecordManager
    sequenced_item_class = SequencedItem
    integer_sequenced_record_class = None
    integer_sequenced_noid_record_class = None
    timestamp_sequenced_record_class = None
    snapshot_record_class = None
    tracking_record_manager_class = None

    def __init__(self, record_manager_class=None, sequenced_item_class=None,
                 integer_sequenced_record_class=None, timestamp_sequenced_record_class=None,
                 tracking_record_manager_class=None,
                 snapshot_record_class=None, contiguous_record_ids=False, application_name=None,
                 pipeline_id=-1):

        self.record_manager_class = record_manager_class or self.record_manager_class

        self.sequenced_item_class = sequenced_item_class or self.sequenced_item_class

        self.integer_sequenced_record_class = integer_sequenced_record_class or self.integer_sequenced_record_class

        self.timestamp_sequenced_record_class = timestamp_sequenced_record_class or \
                                                self.timestamp_sequenced_record_class

        self.tracking_record_manager_class = tracking_record_manager_class or self.tracking_record_manager_class

        self.snapshot_record_class = snapshot_record_class or self.snapshot_record_class

        self.contiguous_record_ids = contiguous_record_ids
        self.application_name = application_name
        self.pipeline_id = pipeline_id

    def construct_integer_sequenced_record_manager(self, record_class=None):
        return self.construct_record_manager(
            record_class=record_class or self.integer_sequenced_record_class,
        )

    def construct_timestamp_sequenced_record_manager(self, record_class=None):
        record_class = record_class or self.timestamp_sequenced_record_class
        assert self.integer_sequenced_record_class
        return self.construct_record_manager(
            record_class=record_class,
        )

    def construct_snapshot_record_manager(self):
        return self.construct_record_manager(
            record_class=self.snapshot_record_class,
        )

    def construct_record_manager(self, record_class, sequenced_item_class=None, **kwargs):
        return self.record_manager_class(
            sequenced_item_class=sequenced_item_class or self.sequenced_item_class,
            record_class=record_class,
            contiguous_record_ids=self.contiguous_record_ids,
            application_name=self.application_name,
            pipeline_id=self.pipeline_id,
            **kwargs
        )

    def construct_tracking_record_manager(self, *args, **kwargs):
        return self.tracking_record_manager_class(*args, **kwargs)

    def construct_datastore(self):
        return None



from math import ceil, log
from random import shuffle
from threading import Thread
from unittest.case import skip
from uuid import UUID, uuid4

from eventsourcing.domain.model.array import AbstractArrayRepository, Array, BigArray
from eventsourcing.exceptions import ArrayIndexError, ConcurrencyError
from eventsourcing.infrastructure.repositories.array import ArrayRepository, BigArrayRepository
from eventsourcing.tests.base import notquick
from eventsourcing.tests.sequenced_item_tests.base import WithPersistencePolicies
from eventsourcing.tests.sequenced_item_tests.test_cassandra_record_manager import \
    WithCassandraRecordManagers
from eventsourcing.tests.sequenced_item_tests.test_django_record_manager import DjangoTestCase
from eventsourcing.tests.sequenced_item_tests.test_sqlalchemy_record_manager import \
    WithSQLAlchemyRecordManagers

try:
    from unittest import mock
except:
    import mock

try:
    from queue import Queue
except ImportError:
    from Queue import Queue


class TestArrayWithSQLAlchemy(WithSQLAlchemyRecordManagers, WithPersistencePolicies):
    def setUp(self):
        super(TestArrayWithSQLAlchemy, self).setUp()
        self.repo = ArrayRepository(
            array_size=3,
            event_store=self.entity_event_store,
        )

    def test_array(self):

        # Start a new array.
        array_id = uuid4()
        array = self.repo[array_id]
        self.assertIsInstance(array, Array)
        self.assertEqual(array.id, array_id)
        self.assertEqual(array, array)
        self.assertNotEqual(array, self.repo[uuid4()])

        # Add some items.
        array[0] = 'item1'
        array.append('item2')
        array.append('item3')

        # Check the array in the self.repo.
        array = self.repo[array_id]

        # Check the array indexing.
        self.assertEqual(array[0], 'item1')
        self.assertEqual(array[-3], 'item1')
        self.assertEqual(array[1], 'item2')
        self.assertEqual(array[-2], 'item2')
        self.assertEqual(array[2], 'item3')
        self.assertEqual(array[-1], 'item3')

        # Check slices also work.
        self.assertEqual(array[0:3], ['item1', 'item2', 'item3'])
        self.assertEqual(array[0:2], ['item1', 'item2'])
        self.assertEqual(array[0:-1], ['item1', 'item2'])
        self.assertEqual(array[0:1], ['item1'])
        self.assertEqual(array[0:-2], ['item1'])
        self.assertEqual(array[1:3], ['item2', 'item3'])
        self.assertEqual(array[1:2], ['item2'])
        self.assertEqual(array[-2:-1], ['item2'])
        self.assertEqual(array[1:1], [])
        self.assertEqual(array[-2:-2], [])
        self.assertEqual(array[2:3], ['item3'])
        self.assertEqual(array[3:3], [])
        self.assertEqual(array[0:300], ['item1', 'item2', 'item3'])
        self.assertEqual(array[2:1], [])
        self.assertEqual(array[2:-2], [])

        self.assertEqual(array[0:], ['item1', 'item2', 'item3'])
        self.assertEqual(array[1:], ['item2', 'item3'])
        self.assertEqual(array[2:], ['item3'])
        self.assertEqual(array[3:], [])
        self.assertEqual(array[4:], [])
        self.assertEqual(array[-1:], ['item3'])
        self.assertEqual(array[-2:], ['item2', 'item3'])
        self.assertEqual(array[-3:], ['item1', 'item2', 'item3'])
        self.assertEqual(array[-4:], ['item1', 'item2', 'item3'])

        self.assertEqual(array[:0], [])
        self.assertEqual(array[:1], ['item1'])
        self.assertEqual(array[:2], ['item1', 'item2'])
        self.assertEqual(array[:3], ['item1', 'item2', 'item3'])
        self.assertEqual(array[:4], ['item1', 'item2', 'item3'])
        self.assertEqual(array[:-1], ['item1', 'item2'])
        self.assertEqual(array[:-2], ['item1'])
        self.assertEqual(array[:-3], [])
        self.assertEqual(array[:-4], [])

        self.assertEqual(array[:], ['item1', 'item2', 'item3'])

        # Check iterator.
        for i, item in enumerate(array):
            self.assertEqual(item, 'item{}'.format(i + 1))

        # Check len.
        self.assertEqual(len(array), 3)

        # Check full error.
        with self.assertRaises(ArrayIndexError):
            # Fail to append fourth item (array size is 3).
            array.append('item1')

        # Check index errors.
        with self.assertRaises(ArrayIndexError):
            # noinspection PyStatementEffect
            array[3]

        with self.assertRaises(ArrayIndexError):
            # noinspection PyStatementEffect
            array[-4]


class BigArrayTestCase(WithSQLAlchemyRecordManagers, WithPersistencePolicies):
    def start_and_append(self, array_size, num_items):
        array = self.get_big_array(array_size)
        items = self.append_items(num_items, array)
        return array, list(items)

    def append_items(self, num_items, array):
        for i in range(num_items):
            item = 'item-{}'.format(i)
            array.append(item)
            yield item

    def start_and_set(self, base_size, num_items, offset=0):
        array = self.get_big_array(base_size)
        items = self.set_items(num_items, array, offset)
        return array, list(items)

    def set_items(self, num_items, array, offset):
        for i in range(offset, offset + num_items):
            item = 'item-{}'.format(i)
            array[i] = item
            yield item

    def get_big_array(self, base_size):
        self.repo.subrepo.array_size = base_size
        root = self.repo[uuid4()]
        return root


class TestBigArrayWithSQLAlchemy(BigArrayTestCase):
    def setUp(self):
        super(TestBigArrayWithSQLAlchemy, self).setUp()
        self.repo = BigArrayRepository(
            array_size=None,
            event_store=self.entity_event_store,
        )
        self.subrepo = self.repo.subrepo
        self.result_queue = None

    def test_big_array_short(self):
        # Can add zero items if base_size is zero.
        root, added = self.start_and_set(0, 0)
        # Check we got a compound array.
        self.assertIsInstance(root, BigArray)
        # Check none was added.
        self.assertEqual(added, [])

        # Can add 1 items if base_size is 1.
        root, added = self.start_and_set(base_size=1, num_items=1)

        # Check we got a compound array.
        self.assertIsInstance(root, BigArray)
        assert isinstance(root, BigArray)
        # Check something was added.
        self.assertTrue(added)

        # Check root has an ID.
        self.assertTrue(root.id)
        # Get the root array.
        root_array = self.subrepo[root.id]
        # Check the root array has next_position 1.
        self.assertEqual(len(root_array), 1)
        # Check the root array is full.
        with self.assertRaises(ArrayIndexError):
            root_array.append(1)

        # Get the current apex ID from the root array.
        apex_id = root_array[-1]
        # Check it's a UUID.
        self.assertIsInstance(apex_id, UUID)
        # Get the apex array.
        apex_array = self.subrepo[apex_id]
        # Check the apex array has next_position 1.
        self.assertEqual(len(apex_array), 1)
        # Check the apex array is full.
        with self.assertRaises(ArrayIndexError):
            apex_array.append(1)

        # Get the item.
        item1 = apex_array[-1]
        # Check it's a UUID.
        self.assertIsInstance(apex_id, UUID)
        # Check it's the last things that was added above.
        self.assertEqual(item1, added[-1])

        # Check the "last array" is the apex we got above.
        last_array, i = root.get_last_array()
        self.assertEqual(last_array, apex_array)
        # Check the 'i' value of the apex is 0.
        self.assertEqual(i, 0)

        # Check the "last item" is the item we got above.
        last_item, next_position = root.get_last_item_and_next_position()
        self.assertEqual(last_item, added[-1])
        # Check the index of the item.
        self.assertEqual(next_position, 1)

        # Can add 2 items if base_size is 2.
        root, added = self.start_and_set(base_size=2, num_items=2)
        last_item, next_position = root.get_last_item_and_next_position()
        self.assertEqual(last_item, added[-1])
        self.assertEqual(next_position, 2)

        # Can add 3 items if base_size is 2.
        root, added = self.start_and_set(base_size=2, num_items=3)
        last_item, next_position = root.get_last_item_and_next_position()
        self.assertEqual(last_item, added[-1])
        self.assertEqual(next_position, 3)
        self.assertEqual(len(self.subrepo[root.id]), 2)
        self.assertEqual(last_array.get_next_position(), 1)

        # Check can append another, that it's not full.
        last_array.append(1)

        # Can add 4 items if base_size is 2.
        root, added = self.start_and_set(base_size=2, num_items=4)
        last_item, next_position = root.get_last_item_and_next_position()
        self.assertEqual(last_item, added[-1])
        self.assertEqual(next_position, 4)

        # Can add 6 items if base_size is 3.
        root, added = self.start_and_set(base_size=3, num_items=7)
        last_item, next_position = root.get_last_item_and_next_position()
        self.assertEqual(last_item, added[-1])
        self.assertEqual(next_position, 7)

        root, added = self.start_and_set(base_size=3, num_items=9)
        last_item, next_position = root.get_last_item_and_next_position()
        self.assertEqual(last_item, added[-1])
        self.assertEqual(next_position, 9)

        root, added = self.start_and_set(base_size=3, num_items=10)
        last_item, next_position = root.get_last_item_and_next_position()
        self.assertEqual(last_item, added[-1])
        self.assertEqual(next_position, 10)

        root, added = self.start_and_set(base_size=4, num_items=16)
        last_item, next_position = root.get_last_item_and_next_position()
        self.assertEqual(last_item, added[-1])
        self.assertEqual(next_position, 16)

        root, added = self.start_and_set(base_size=4, num_items=27)
        last_item, next_position = root.get_last_item_and_next_position()
        self.assertEqual(last_item, added[-1])
        self.assertEqual(next_position, 27)

        # Can't add 2 items if base_size is 1.
        with self.assertRaises(ArrayIndexError):
            self.start_and_set(base_size=1, num_items=2)

        # Can't add 5 items if base_size is 2.
        with self.assertRaises(ArrayIndexError):
            self.start_and_set(base_size=2, num_items=5)

        # Can't add 28 items if base_size is 3.
        with self.assertRaises(ArrayIndexError):
            self.start_and_set(base_size=3, num_items=28)

    @notquick
    def test_big_array_longer(self):
        # Can add 256 items if base_size is 4.
        root, added = self.start_and_set(base_size=4, num_items=256)

        self.assertEqual(len(list(self.repo[root.id][:])), 256)
        # self.assertEqual(len(self.repo[root.id]), 256)

        # Can't add 257 items if base_size is 4.
        with self.assertRaises(ArrayIndexError):
            self.start_and_set(base_size=4, num_items=257)

        # Can add 156 items if base_size is 4.
        root, added = self.start_and_set(base_size=4, num_items=156)

        self.assertEqual(len(list(self.repo[root.id][:])), 256)
        self.assertEqual(self.repo[root.id].get_next_position(), 156)

        # Can add 100 items if base_size is 10000.
        #  - Should have capacity for 10000**10000 items,
        #    which is 1e+40000 items, but is not checked here.
        root, added = self.start_and_set(base_size=10000, num_items=100)
        self.assertEqual(root.get_last_item_and_next_position(), (added[-1], 100))
        # Check depth is 1.
        self.assertEqual(self.subrepo[root.id].get_next_position(), 1)

        # Can add 101 items if array_size is 100.
        root, added = self.start_and_set(base_size=100, num_items=101)
        self.assertEqual(root.get_last_item_and_next_position(), (added[-1], 101))
        # Check depth is 2.
        self.assertEqual(self.subrepo[root.id].get_last_item_and_next_position()[1], 2)

    def test_big_array_bigoffset(self):
        # Can add 256 items with large offset.
        base_size = 4000
        offset = 400000000000000000000000000000000000
        root, added = self.start_and_set(base_size=base_size, num_items=256, offset=offset)
        self.assertEqual(self.subrepo[root.id].get_next_position(), 10)  # depth of array tree
        self.assertEqual(len(list(self.repo[root.id][offset:offset + 256])), 256)

        # Get items at the start of the array (all should be None).
        self.assertEqual(len(list(self.repo[root.id][0:10])), 10)
        self.assertEqual(len(list(self.repo[root.id][0:10000])), 10000)

    @notquick
    def _test_big_array_biggest_offset(self):
        # Can add 25 items at the end.
        base_size = 4000
        offset = (base_size ** base_size) - 25
        root, added = self.start_and_set(base_size=base_size, num_items=25, offset=offset)
        self.assertEqual(self.subrepo[root.id].get_last_item_and_next_position()[1], base_size)  # depth of array tree
        self.assertEqual(len(list(self.repo[root.id][offset:])), 25)

    @notquick
    def test_big_array_too_much_offset(self):
        # Can't add 1 item past the end.
        base_size = 4000
        offset = (base_size ** base_size)
        with self.assertRaises(ArrayIndexError):
            self.start_and_set(base_size=base_size, num_items=1, offset=offset)

    def test_iterator_2_3(self):
        self._test_iterator(array_size=2, num_items=3)

    def test_iterator_2_4(self):
        self._test_iterator(array_size=2, num_items=3)

    def test_iterator_3_20(self):
        self._test_iterator(array_size=3, num_items=20)

    def test_iterator_3_27(self):
        self._test_iterator(array_size=3, num_items=27)

    @notquick
    def test_iterator_4_230(self):
        self._test_iterator(array_size=4, num_items=230)

    def test_iterator_1000_23(self):
        self._test_iterator(array_size=1000, num_items=23)

    def _test_iterator(self, array_size, num_items):
        array, added = self.start_and_set(array_size, num_items)

        self.assertEqual(array[0], added[0])
        self.assertEqual(array[-3], added[-3])
        self.assertEqual(array[1], added[1])
        self.assertEqual(array[-2], added[-2])
        self.assertEqual(array[2], added[2])
        self.assertEqual(array[-1], added[-1])

        # Check slices also work.
        self.assertEqual(list(array[0:3]), added[0:3])
        self.assertEqual(list(array[0:2]), added[0:2])
        self.assertEqual(list(array[0:1]), added[0:1])
        self.assertEqual(list(array[1:3]), added[1:3])
        self.assertEqual(list(array[1:2]), added[1:2])
        self.assertEqual(list(array[1:1]), added[1:1])
        self.assertEqual(list(array[2:3]), added[2:3])
        self.assertEqual(list(array[3:3]), added[3:3])
        self.assertEqual(list(array[2:1]), added[2:1])

        if array_size < 10:

            if num_items == array_size ** array_size:
                self.assertEqual(list(array[0:300]), added[0:300])
                self.assertEqual(list(array[-2:-1]), added[-2:-1])
                self.assertEqual(list(array[-2:-2]), added[-2:-2])
                self.assertEqual(list(array[0:-1]), added[0:-1])
                self.assertEqual(list(array[0:-2]), added[0:-2])

                self.assertEqual(list(array[-1:]), added[-1:])
                self.assertEqual(list(array[-2:]), added[-2:])
                self.assertEqual(list(array[-3:]), added[-3:])
                self.assertEqual(list(array[-4:]), added[-4:])

                self.assertEqual(list(array[2:-2]), added[2:-2])

                self.assertEqual(list(array[0:]), added[0:])
                self.assertEqual(list(array[1:]), added[1:])
                self.assertEqual(list(array[2:]), added[2:])
                self.assertEqual(list(array[3:]), added[3:])
                self.assertEqual(list(array[4:]), added[4:])
                self.assertEqual(list(array[:0]), added[:0])
                self.assertEqual(list(array[:1]), added[:1])
                self.assertEqual(list(array[:2]), added[:2])
                self.assertEqual(list(array[:3]), added[:3])
                self.assertEqual(list(array[:4]), added[:4])
                self.assertEqual(list(array[:-1]), added[:-1])
                self.assertEqual(list(array[:-2]), added[:-2])
                self.assertEqual(list(array[:-3]), added[:-3])
                self.assertEqual(list(array[:-4]), added[:-4])

                # Check len.
                self.assertEqual(len(array), num_items)

            # Check iterator.
            for i, item in enumerate(array):
                self.assertEqual(item, 'item-{}'.format(i))

        # Check index errors.
        # - out of range
        with self.assertRaises(ArrayIndexError):
            # noinspection PyStatementEffect
            array[num_items]

        with self.assertRaises(ArrayIndexError):
            # noinspection PyStatementEffect
            array[- num_items - 1]

        return array, added

    def test_setitem(self):

        a = self.get_big_array(1)
        # self.assertEqual(a.repo.array_size, 1)
        self.assertEqual(list(a[:]), [None])
        a[0] = 'item-0'
        self.assertEqual(list(a[:]), ['item-0'])
        with self.assertRaises(ConcurrencyError):
            a[0] = 'item-0'

        a = self.get_big_array(2)
        a[0] = 'item-0'
        self.assertEqual(list(a[:]), ['item-0', None, None, None])
        a[1] = 'item-1'
        self.assertEqual(list(a[:]), ['item-0', 'item-1', None, None])
        a[2] = 'item-2'
        self.assertEqual(list(a[:]), ['item-0', 'item-1', 'item-2', None])

        # Add four items in order to an big array with array size 4.
        a = self.get_big_array(4)

        assert isinstance(a, BigArray)
        items = []
        for i in range(4):
            item = 'item-{}'.format(i)
            a[i] = item
            items.append(item)

        self.assertEqual(list(a[:4]), items)

        # Add three items in reverse order to big array with array size 4.
        a = self.get_big_array(4)
        items = []
        for i in reversed(range(1, 4)):
            item = 'item-{}'.format(i)
            a[i] = item
            items.append(item)

        self.assertEqual(list(a[1:4]), list(reversed(items)))

        # Check getting the first item (which wasn't set) raises IndexError.
        with self.assertRaises(ArrayIndexError):
            a[0]

        item0 = 'item-0'
        a[0] = item0
        self.assertEqual(a[0], item0)
        items.append(item0)

        # Add fifth item, which should trigger extending the compound
        # by one base array.
        a[4] = 'item-4'
        self.assertEqual(list(a[0:5]), list(reversed(items)) + ['item-4'])

        # Add sixth item, which should trigger extending the compound
        # by one base array.
        a[40] = 'item-40'
        self.assertEqual(list(a[0:41]), list(reversed(items)) + ['item-4'] + [None] * 35 + ['item-40'])

        a[20] = 'item-20'
        self.assertEqual(list(a[0:41]), list(reversed(items)) + ['item-4'] + [None] * 15 + ['item-20'] + [None] * 19
                         + ['item-40'])

        # Add item at position 200, which should trigger extending
        # the compound by several base arrays. Building a full
        # tree would be expensive, so maybe just pick out the
        # single path to the root, adding in the things in expected
        # positions, so "last" can be quite far quite quickly.

        # Add 20 items in reverse order to big array with array size 4.
        a = self.get_big_array(4)
        items = []
        for i in reversed(range(20)):
            item = 'item-{}'.format(i)
            a[i] = item
            items.append(item)

        self.assertEqual(list(a[:20]), list(reversed(items)))

        # Add 20 items in shuffled order to big array with array size 4.
        a = self.get_big_array(3)
        items = []
        indexes = list(range(11))
        shuffle(indexes)
        for i in indexes:
            item = 'item-{}'.format(i)
            a[i] = item
            items.append(item)

        self.assertEqual(list(sorted(a[:11])), list(sorted(reversed(items))))

    def test_calc_required_height(self):
        root = BigArray(
            array_id='1',
            repo=mock.Mock(spec=AbstractArrayRepository),
        )
        with self.assertRaises(ValueError):
            root.calc_required_height(0, 0)
        self.assertEqual(root.calc_required_height(0, 1), 1)
        with self.assertRaises(ArrayIndexError):
            root.calc_required_height(1, 1)
        self.assertEqual(root.calc_required_height(0, 2), 1)
        self.assertEqual(root.calc_required_height(1, 2), 1)
        self.assertEqual(root.calc_required_height(2, 2), 2)
        self.assertEqual(root.calc_required_height(3, 2), 2)
        with self.assertRaises(ArrayIndexError):
            root.calc_required_height(4, 2)

        self.assertEqual(root.calc_required_height(0, 3), 1)
        self.assertEqual(root.calc_required_height(1, 3), 1)
        self.assertEqual(root.calc_required_height(2, 3), 1)
        self.assertEqual(root.calc_required_height(3, 3), 2)
        self.assertEqual(root.calc_required_height(4, 3), 2)
        self.assertEqual(root.calc_required_height(5, 3), 2)
        self.assertEqual(root.calc_required_height(6, 3), 2)
        self.assertEqual(root.calc_required_height(7, 3), 2)
        self.assertEqual(root.calc_required_height(8, 3), 2)
        self.assertEqual(root.calc_required_height(9, 3), 3)
        self.assertEqual(root.calc_required_height(10, 3), 3)
        self.assertEqual(root.calc_required_height(11, 3), 3)
        self.assertEqual(root.calc_required_height(12, 3), 3)
        self.assertEqual(root.calc_required_height(13, 3), 3)
        self.assertEqual(root.calc_required_height(14, 3), 3)
        self.assertEqual(root.calc_required_height(15, 3), 3)
        self.assertEqual(root.calc_required_height(16, 3), 3)
        self.assertEqual(root.calc_required_height(17, 3), 3)
        self.assertEqual(root.calc_required_height(18, 3), 3)
        self.assertEqual(root.calc_required_height(19, 3), 3)
        self.assertEqual(root.calc_required_height(20, 3), 3)
        self.assertEqual(root.calc_required_height(21, 3), 3)
        self.assertEqual(root.calc_required_height(22, 3), 3)
        self.assertEqual(root.calc_required_height(23, 3), 3)
        self.assertEqual(root.calc_required_height(24, 3), 3)
        self.assertEqual(root.calc_required_height(25, 3), 3)
        self.assertEqual(root.calc_required_height(26, 3), 3)
        with self.assertRaises(ArrayIndexError):
            root.calc_required_height(27, 3)

        self.assertEqual(root.calc_required_height(100 ** 100 - 1, 100), 100)
        self.assertEqual(root.calc_required_height(400 ** 400 - 1, 400), 400)
        self.assertEqual(root.calc_required_height(900 ** 900 - 1, 900), 900)
        self.assertEqual(root.calc_required_height(999 ** 999 - 1, 999), 999)
        self.assertEqual(root.calc_required_height(1000 ** 1000 - 1, 1000), 1000)
        self.assertEqual(root.calc_required_height(1001 ** 1001 - 1, 1001), 1001)
        self.assertEqual(root.calc_required_height(1100 ** 1100 - 1, 1100), 1100)
        self.assertEqual(root.calc_required_height(2000 ** 2000 - 1, 2000), 2000)
        self.assertEqual(root.calc_required_height(10000 ** 10000 - 1, 10000), 10000)


class TestBigArrayWithSQLAlchemyAndMultithreading(BigArrayTestCase):
    use_named_temporary_file = True

    def setUp(self):
        super(TestBigArrayWithSQLAlchemyAndMultithreading, self).setUp()
        self.repo = BigArrayRepository(
            event_store=self.entity_event_store,
        )
        self.subrepo = self.repo.subrepo
        self.result_queue = None

    def test_big_array_threads_1_1_1(self):
        self._test_big_array_threads(1, 1, 1)

    def test_big_array_threads_2_2_2(self):
        self._test_big_array_threads(2, 2, 2)

    def test_big_array_threads_3_3_9(self):
        self._test_big_array_threads(3, 3, 9)

    def test_big_array_threads_3_9_3(self):
        self._test_big_array_threads(3, 9, 3)

    @notquick
    @skip("Avoid 'database is locked' error from SQLite")
    def test_big_array_threads_4_4_64(self):
        self._test_big_array_threads(4, 4, 64)

    @notquick
    @skip("Avoid 'database is locked' error from SQLite")
    def test_big_array_threads_4_8_32(self):
        self._test_big_array_threads(4, 8, 32)

    @notquick
    @skip("Avoid 'database is locked' error from SQLite")
    def test_big_array_threads_4_16_16(self):
        self._test_big_array_threads(4, 16, 16)

    @notquick
    @skip("Avoid 'database is locked' error from SQLite")
    def test_big_array_threads_4_32_8(self):
        self._test_big_array_threads(4, 32, 8)

    @notquick
    @skip("Avoid 'database is locked' error from SQLite")
    def test_big_array_threads_4_64_4(self):
        self._test_big_array_threads(4, 64, 4)

    @notquick
    @skip("Avoid 'database is locked' error from SQLite")
    def test_big_array_threads_4_256_1(self):
        self._test_big_array_threads(4, 256, 1)

    def _test_big_array_threads(self, array_size, num_threads, num_items_per_thread):
        error_queue = Queue()
        full_queue = Queue()
        self.result_queue = Queue()
        assert num_threads * num_items_per_thread == array_size ** array_size, (
            num_threads * num_items_per_thread, array_size ** array_size)
        root = self.get_big_array(base_size=array_size)

        def task():
            try:
                for item in self.append_items(num_items_per_thread, root):
                    self.result_queue.put(item)
            except ArrayIndexError as e:
                full_queue.put(e)
            except Exception as e:
                error_queue.put(e)

        # Have many threads each trying to fill up the log
        threads = [Thread(target=task) for _ in range(num_threads)]
        for t in threads:
            t.start()

        for t in threads:
            t.join()

        # Gather thread exception.
        errors = []
        while not error_queue.empty():
            errors.append(error_queue.get())

        # Gather thread exception.
        fulls = []
        while not full_queue.empty():
            fulls.append(full_queue.get())

        results = []
        while not self.result_queue.empty():
            results.append(self.result_queue.get())

        # Check there are no thread exceptions.
        if errors:
            raise errors[0]
        self.assertFalse(errors)

        if fulls:
            raise fulls[0]
        self.assertFalse(fulls)

        expected_results = min(array_size ** array_size, num_threads * num_items_per_thread)
        self.assertEqual(len(results), expected_results)

        # Check the height of the root.
        # - limited either by the max capacity
        #   or by the number of added item

        if expected_results <= 1:
            expected_height = expected_results
        else:
            expected_height = ceil(log(expected_results, array_size))

        self.assertEqual(self.subrepo[root.id].get_next_position(), expected_height)


class TestArrayWithCassandra(WithCassandraRecordManagers, TestArrayWithSQLAlchemy):
    pass


class TestArrayWithDjango(DjangoTestCase, TestArrayWithSQLAlchemy):
    pass


class TestBigArrayWithCassandra(WithCassandraRecordManagers, TestBigArrayWithSQLAlchemy):
    pass


class TestBigArrayWithDjango(DjangoTestCase, TestBigArrayWithSQLAlchemy):
    pass


class TestBigArrayWithCassandraAndMultithreading(WithCassandraRecordManagers,
                                                 TestBigArrayWithSQLAlchemyAndMultithreading):
    pass



# Generated by Django 2.0 on 2018-01-21 18:25

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='IntegerSequencedRecord',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False)),
                ('sequence_id', models.UUIDField()),
                ('position', models.BigIntegerField()),
                ('topic', models.TextField()),
                ('data', models.TextField()),
            ],
            options={
                'db_table': 'integer_sequenced_items',
            },
        ),
        migrations.CreateModel(
            name='SnapshotRecord',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False)),
                ('sequence_id', models.UUIDField()),
                ('position', models.BigIntegerField()),
                ('topic', models.TextField()),
                ('data', models.TextField()),
            ],
            options={
                'db_table': 'snapshots',
            },
        ),
        migrations.CreateModel(
            name='StoredEventRecord',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False)),
                ('originator_id', models.UUIDField()),
                ('originator_version', models.BigIntegerField()),
                ('event_type', models.TextField()),
                ('state', models.TextField()),
            ],
            options={
                'db_table': 'stored_events',
            },
        ),
        migrations.CreateModel(
            name='TimestampSequencedRecord',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False)),
                ('sequence_id', models.UUIDField()),
                ('position', models.DecimalField(decimal_places=6, max_digits=24)),
                ('topic', models.TextField()),
                ('data', models.TextField()),
            ],
            options={
                'db_table': 'timestamp_sequenced_items',
            },
        ),
        migrations.AddIndex(
            model_name='timestampsequencedrecord',
            index=models.Index(fields=['position'], name='position_idx'),
        ),
        migrations.AlterUniqueTogether(
            name='timestampsequencedrecord',
            unique_together={('sequence_id', 'position')},
        ),
        migrations.AlterUniqueTogether(
            name='storedeventrecord',
            unique_together={('originator_id', 'originator_version')},
        ),
        migrations.AlterUniqueTogether(
            name='snapshotrecord',
            unique_together={('sequence_id', 'position')},
        ),
        migrations.AlterUniqueTogether(
            name='integersequencedrecord',
            unique_together={('sequence_id', 'position')},
        ),
    ]



import six
from django.db import IntegrityError, connection, transaction

from eventsourcing.infrastructure.base import RelationalRecordManager


class DjangoRecordManager(RelationalRecordManager):
    def __init__(self, *args, **kwargs):
        super(DjangoRecordManager, self).__init__(*args, **kwargs)

    def write_records(self, records, tracking_kwargs=None):
        try:
            with transaction.atomic(self.record_class.objects.db):
                if self.contiguous_record_ids:
                    # Use cursor to execute insert select max statement.
                    with connection.cursor() as cursor:
                        for record in records:
                            # Get values from record obj.
                            params = []
                            for col_name in self.field_names:
                                col_value = getattr(record, col_name)
                                col_type = self.record_class._meta.get_field(col_name)

                                # Prepare value for database.
                                param = col_type.get_db_prep_value(col_value, connection)
                                params.append(param)

                            # Execute insert statement.
                            cursor.execute(self.insert_select_max, params)
                else:
                    # Todo: If it's faster, change to use an "insert_values" raw query.
                    # Save record objects.
                    for record in records:
                        record.save()

        except IntegrityError as e:
            self.raise_record_integrity_error(e)

    def _prepare_insert(self, tmpl, record_class, field_names, placeholder_for_id=False):
        """
        With transaction isolation level of "read committed" this should
        generate records with a contiguous sequence of integer IDs, using
        an indexed ID column, the database-side SQL max function, the
        insert-select-from form, and optimistic concurrency control.
        """
        statement = tmpl.format(
            tablename=self.get_record_table_name(record_class),
            columns=", ".join(field_names),
            placeholders=", ".join(['%s' for _ in field_names]),
        )
        return statement

    def get_record_table_name(self, record_class):
        """Returns table name from SQLAlchemy record class."""
        return record_class._meta.db_table

    def get_item(self, sequence_id, position):
        records = self.record_class.objects.filter(sequence_id=sequence_id, position=position).all()
        return self.from_record(records[0])

    def get_records(self, sequence_id, gt=None, gte=None, lt=None, lte=None, limit=None,
                    query_ascending=True, results_ascending=True):

        assert limit is None or limit >= 1, limit

        filter_kwargs = {self.field_names.sequence_id: sequence_id}
        query = self.record_class.objects.filter(**filter_kwargs)

        position_field_name = self.field_names.position

        if query_ascending:
            query = query.order_by(position_field_name)
        else:
            query = query.order_by('-' + position_field_name)

        if gt is not None:
            arg = '{}__gt'.format(position_field_name)
            query = query.filter(**{arg: gt})
        if gte is not None:
            arg = '{}__gte'.format(position_field_name)
            query = query.filter(**{arg: gte})
        if lt is not None:
            arg = '{}__lt'.format(position_field_name)
            query = query.filter(**{arg: lt})
        if lte is not None:
            arg = '{}__lte'.format(position_field_name)
            query = query.filter(**{arg: lte})

        if limit is not None:
            query = query[:limit]

        records = query.all()

        if results_ascending != query_ascending:
            # This code path is under test, but not otherwise used ATM.
            records = list(records)
            records.reverse()

        return records

    def get_notifications(self, start=None, stop=None, *args, **kwargs):
        """
        Returns all records in the table.
        """
        filter_kwargs = {}
        position_field_name = 'id'
        # Todo: Also support sequencing by 'position' if items are sequenced by timestamp?
        if start is not None:
            filter_kwargs['%s__gte' % position_field_name] = start + 1
        if stop is not None:
            filter_kwargs['%s__lt' % position_field_name] = stop + 1
        query = self.record_class.objects.filter(**filter_kwargs)
        query = query.order_by('%s' % position_field_name)
        return query.all()

    def delete_record(self, record):
        """
        Permanently removes record from table.
        """
        record.delete()

    def get_max_record_id(self):
        try:
            return self.record_class.objects.latest('id').id
        except self.record_class.DoesNotExist:
            return None

    def all_sequence_ids(self):
        sequence_id_fieldname = self.field_names.sequence_id
        values_queryset = self.record_class.objects.values(sequence_id_fieldname).distinct()
        for values in values_queryset:
            yield values[sequence_id_fieldname]



from uuid import UUID

from eventsourcing.application.policies import PersistencePolicy
from eventsourcing.domain.model.events import DomainEvent
from eventsourcing.example.domainmodel import create_new_example
from eventsourcing.example.infrastructure import ExampleRepository
from eventsourcing.infrastructure.cassandra.records import StoredEventRecord
from eventsourcing.infrastructure.cassandra.manager import CassandraRecordManager
from eventsourcing.infrastructure.cassandra.datastore import CassandraDatastore, CassandraSettings
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.sequenceditem import StoredEvent
from eventsourcing.infrastructure.sequenceditemmapper import SequencedItemMapper
from eventsourcing.tests.datastore_tests.base import AbstractDatastoreTestCase


# In this test the default SequencedItem class is replaced with a "stored event" class.
# How easy is it to customize the infrastructure to support that? We just need
# to define the new sequenced item class, define a suitable record class,
# and configure the other components. It's easy.

class ExampleApplicationWithAlternativeSequencedItemType(object):
    def __init__(self):
        self.event_store = EventStore(
            record_manager=CassandraRecordManager(
                record_class=StoredEventRecord,
                sequenced_item_class=StoredEvent,
            ),
            sequenced_item_mapper=SequencedItemMapper(
                sequenced_item_class=StoredEvent,
                other_attr_names=(),
            )
        )
        self.repository = ExampleRepository(
            event_store=self.event_store,
        )
        self.persistence_policy = PersistencePolicy(
            event_store=self.event_store,
            event_type=DomainEvent
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.persistence_policy.close()


class TestExampleWithAlternativeSequencedItemType(AbstractDatastoreTestCase):
    def setUp(self):
        super(TestExampleWithAlternativeSequencedItemType, self).setUp()
        self.datastore.setup_connection()
        self.datastore.setup_tables()

    def tearDown(self):
        self.datastore.drop_tables()
        self.datastore.close_connection()
        super(TestExampleWithAlternativeSequencedItemType, self).setUp()

    def construct_datastore(self):
        return CassandraDatastore(
            settings=CassandraSettings(),
            tables=(StoredEventRecord,)

        )

    def test(self):
        with ExampleApplicationWithAlternativeSequencedItemType() as app:
            # Create entity.
            entity1 = create_new_example(a='a', b='b')
            self.assertIsInstance(entity1.id, UUID)
            self.assertEqual(entity1.a, 'a')
            self.assertEqual(entity1.b, 'b')

            # Check there is a stored event.
            all_records = list(app.event_store.record_manager.get_records(entity1.id))
            assert len(all_records) == 1, len(all_records)
            stored_event = all_records[0]
            assert isinstance(stored_event, StoredEventRecord), stored_event
            assert stored_event.originator_id == entity1.id
            assert stored_event.originator_version == 0

            # Read entity from repo.
            retrieved_obj = app.repository[entity1.id]
            self.assertEqual(retrieved_obj.id, entity1.id)



from eventsourcing.utils.cipher.aes import AESCipher
from eventsourcing.tests.example_application_tests.base import WithExampleApplication
from eventsourcing.tests.example_application_tests.test_example_application_with_cassandra import \
    TestExampleApplicationWithCassandra
from eventsourcing.tests.example_application_tests.test_example_application_with_sqlalchemy import \
    TestExampleApplicationWithSQLAlchemy


class WithEncryption(WithExampleApplication):
    def construct_cipher(self):
        return AESCipher(cipher_key=b'0123456789abcdef')


class TestEncryptedApplicationWithCassandra(WithEncryption, TestExampleApplicationWithCassandra):
    pass


class TestEncryptedApplicationWithSQLAlchemy(WithEncryption, TestExampleApplicationWithSQLAlchemy):
    pass



from cassandra.cqlengine.models import columns, Model


class IntegerSequencedRecord(Model):
    """Stores integer-sequenced items in Cassandra."""
    __table_name__ = 'integer_sequenced_items'
    _if_not_exists = True

    # Sequence ID (e.g. an entity or aggregate ID).
    sequence_id = columns.UUID(partition_key=True)

    # Position (index) of item in sequence.
    position = columns.BigInt(clustering_order='DESC', primary_key=True)

    # Topic of the item (e.g. path to domain event class).
    topic = columns.Text(required=True)

    # State of the item (serialized dict, possibly encrypted).
    data = columns.Text(required=True)


class TimestampSequencedRecord(Model):
    """Stores timestamp-sequenced items in Cassandra."""
    __table_name__ = 'timestamp_sequenced_items'
    _if_not_exists = True

    # Sequence ID (e.g. an entity or aggregate ID).
    sequence_id = columns.UUID(partition_key=True)

    # Position (in time) of item in sequence.
    position = columns.Decimal(clustering_order='DESC', primary_key=True)

    # Topic of the item (e.g. path to domain event class).
    topic = columns.Text(required=True)

    # State of the item (serialized dict, possibly encrypted).
    data = columns.Text(required=True)


class TimeuuidSequencedRecord(Model):
    """Stores timeuuid-sequenced items in Cassandra."""
    __table_name__ = 'timeuuid_sequenced_items'
    _if_not_exists = True

    # Sequence UUID (e.g. an entity or aggregate ID).
    sequence_id = columns.UUID(partition_key=True)

    # Position (in time) of item in sequence.
    position = columns.TimeUUID(clustering_order='DESC', primary_key=True)

    # Topic of the item (e.g. path to domain event class).
    topic = columns.Text(required=True)

    # State of the item (serialized dict, possibly encrypted).
    data = columns.Text(required=True)


class SnapshotRecord(Model):
    """Stores snapshots in Cassandra."""
    __table_name__ = 'snapshots'
    _if_not_exists = True

    # Sequence ID (e.g. an entity or aggregate ID).
    sequence_id = columns.UUID(partition_key=True)

    # Position (index) of item in sequence.
    position = columns.BigInt(clustering_order='DESC', primary_key=True)

    # Topic of the item (e.g. path to domain entity class).
    topic = columns.Text(required=True)

    # State of the entity (serialized dict, possibly encrypted).
    data = columns.Text(required=True)


class StoredEventRecord(Model):
    """Stores integer-sequenced items in Cassandra."""
    __table_name__ = 'stored_events'
    _if_not_exists = True

    # Aggregate ID (e.g. an entity or aggregate ID).
    originator_id = columns.UUID(partition_key=True)

    # Aggregate version (index) of item in sequence.
    originator_version = columns.BigInt(clustering_order='DESC', primary_key=True)

    # Topic of the item (e.g. path to domain event class).
    event_type = columns.Text(required=True)

    # State of the item (serialized dict, possibly encrypted).
    state = columns.Text(required=True)



# # coding=utf-8
# from __future__ import unicode_literals
#
# import os
#
# from eventsourcing.contrib.suffixtrees.domain.model.suffixtree import SuffixTree, SuffixTreeApplication, \
#     register_new_suffix_tree
# from eventsourcing.tests.base import notquick
# from eventsourcing.tests.sequenced_item_tests.base import WithPersistencePolicy
# from eventsourcing.tests.unit_test_fixtures_suffix_tree_text import LONG_TEXT
#
# LONG_TEXT_FIXTURE_PATH = os.path.join(os.path.dirname(__file__), 'test_suffix_tree.txt')
#
#
# class SuffixTreeTest(WithPersistencePolicy):
#
#     def setUp(self):
#         self.app = SuffixTreeApplication(
#             entity_record_manager=self.integer_sequence_record_manager,
#         )
#
#     def tearDown(self):
#         self.app.close()
#
#     def test_empty_string(self):
#         st = register_new_suffix_tree()
#         assert isinstance(st, SuffixTree)
#         st.add_string('')
#         self.assertEqual(self.app.find_substring('not there', st.id), -1)
#         self.assertEqual(self.app.find_substring('', st.id), -1)
#         self.assertFalse(self.app.has_substring('not there', st.id))
#         self.assertFalse(self.app.has_substring('', st.id))
#
#     def test_repeated_string(self):
#         st = register_new_suffix_tree()
#         st.add_string("aaa")
#         self.assertEqual(self.app.find_substring('a', st.id), 0)
#         self.assertEqual(self.app.find_substring('aa', st.id), 0)
#         self.assertEqual(self.app.find_substring('aaa', st.id), 0)
#         self.assertEqual(self.app.find_substring('b', st.id), -1)
#         self.assertTrue(self.app.has_substring('a', st.id))
#         self.assertTrue(self.app.has_substring('aa', st.id))
#         self.assertTrue(self.app.has_substring('aaa', st.id))
#         self.assertFalse(self.app.has_substring('aaaa', st.id))
#         self.assertFalse(self.app.has_substring('b', st.id))
#         # case sensitive by default
#         self.assertFalse(self.app.has_substring('A', st.id))
#
#     def test_mississippi(self):
#         st = register_new_suffix_tree()
#         st.add_string("mississippi")
#         self.assertEqual(self.app.find_substring('a', st.id), -1)
#         self.assertEqual(self.app.find_substring('m', st.id), 0)
#         self.assertEqual(self.app.find_substring('i', st.id), 1)
#
#     @notquick
#     def test_long_string(self):
#         st = register_new_suffix_tree()
#         st.add_string(LONG_TEXT)
#         self.assertEqual(self.app.find_substring('Ukkonen', st.id), 1498)
#         self.assertEqual(self.app.find_substring('Optimal', st.id), 11074)
#         self.assertFalse(self.app.has_substring('ukkonen', st.id))
#
#     @notquick
#     def test_case_insensitivity(self):
#         st = register_new_suffix_tree(case_insensitive=True)
#         st.add_string(LONG_TEXT)
#         self.assertEqual(self.app.find_substring('ukkonen', st.id), 1498)
#         self.assertEqual(self.app.find_substring('Optimal', st.id), 1830)



import unittest
from time import time
from uuid import UUID, uuid4, uuid1

from decimal import Decimal

from eventsourcing.domain.model.decorators import subscribe_to
from eventsourcing.domain.model.events import DomainEvent, EventHandlersNotEmptyError, EventWithOriginatorID, \
    EventWithOriginatorVersion, EventWithTimestamp, _event_handlers, assert_event_handlers_empty, \
    create_timesequenced_event_id, publish, subscribe, unsubscribe, EventWithTimeuuid
from eventsourcing.utils.topic import resolve_topic, get_topic
from eventsourcing.example.domainmodel import Example
from eventsourcing.exceptions import TopicResolutionError
from eventsourcing.utils.times import decimaltimestamp_from_uuid, decimaltimestamp

try:
    from unittest import mock
except ImportError:
    import mock


class Event(DomainEvent):
    pass


# Check not equal to different type with same values.
class SubclassEvent(Event):
    pass


class TestAbstractDomainEvent(unittest.TestCase):

    def test(self):
        # Check base class can be sub-classed.

        # Check subclass can be instantiated.
        event1 = Event()
        self.assertEqual(type(event1).__qualname__, 'Event')

        # Check subclass can be instantiated with other parameters.
        event2 = Event(name='value')

        # Check the attribute value is available.
        self.assertEqual(event2.name, 'value')

        # Check the attribute value cannot be changed
        with self.assertRaises(AttributeError):
            event2.name = 'another value'
        self.assertEqual(event2.name, 'value')

        # Check it's equal to itself, by value.
        self.assertEqual(event2, event2)
        self.assertEqual(event2, Event(name='value'))

        # Check not equal to same event type with different values.
        self.assertNotEqual(event2, Event(name='another value'))

        self.assertNotEqual(event2, SubclassEvent(name=event2.name))


class TestEventWithOriginatorID(unittest.TestCase):
    def test(self):
        # Check base class can be sub-classed.
        class Event(EventWithOriginatorID):
            pass

        # Check can't instantiate without an ID.
        with self.assertRaises(TypeError):
            # noinspection PyArgumentList
            Event()

        # Check any kind of ID is acceptable.
        event = Event(originator_id='1')
        self.assertEqual(event.originator_id, '1')

        event = Event(originator_id=1)
        self.assertEqual(event.originator_id, 1)

        event = Event(originator_id=uuid4())
        self.assertIsInstance(event.originator_id, UUID)

        # Check the ID value can't be reassigned.
        with self.assertRaises(AttributeError):
            # noinspection PyPropertyAccess
            event.originator_id = '2'


class TestEventWithOriginatorVersion(unittest.TestCase):
    def test(self):
        # Check base class can be sub-classed.
        class Event(EventWithOriginatorVersion):
            pass

        # Check event can be instantiated with a version.
        version1 = 1
        event = Event(originator_version=version1)
        self.assertEqual(event.originator_version, version1)

        # Check event can't be instantiated without a version.
        with self.assertRaises(TypeError):
            # noinspection PyArgumentList
            event = Event()

        # Check version must be an integer.
        with self.assertRaises(TypeError):
            event = Event(originator_version='1')

        # Check the version value can't be reassigned.
        with self.assertRaises(AttributeError):
            # noinspection PyPropertyAccess
            event.originator_version = 2


class TestEventWithTimestamp(unittest.TestCase):
    def test(self):
        # Check base class can be sub-classed.
        class Event(EventWithTimestamp):
            pass

        # Check event can be instantiated with a timestamp.
        time1 = decimaltimestamp()
        event = Event(timestamp=time1)
        self.assertEqual(event.timestamp, time1)

        # Check event can be instantiated without a timestamp.
        event = Event()
        self.assertGreater(event.timestamp, time1)
        self.assertLess(event.timestamp, decimaltimestamp())

        # Check the timestamp value can't be reassigned.
        with self.assertRaises(AttributeError):
            # noinspection PyPropertyAccess
            event.timestamp = decimaltimestamp()


class TestEventWithTimeuuid(unittest.TestCase):
    def test(self):
        # Check base class can be sub-classed.
        class Event(EventWithTimeuuid):
            pass

        # Check event can be instantiated with an event_id.
        event_id = uuid1()
        event = Event(event_id=event_id)
        self.assertEqual(event.event_id, event_id)

        # Check event can be instantiated without an event_id.
        time1 = decimaltimestamp()
        event = Event()
        self.assertGreater(decimaltimestamp_from_uuid(event.event_id), time1)
        self.assertLess(decimaltimestamp_from_uuid(event.event_id), decimaltimestamp())

        # Check the event_id can't be reassigned.
        with self.assertRaises(AttributeError):
            # noinspection PyPropertyAccess
            event.event_id = decimaltimestamp()


class TestEventWithOriginatorVersionAndID(unittest.TestCase):
    # noinspection PyArgumentList
    def test(self):
        # Check base class can be sub-classed.
        class Event(EventWithOriginatorVersion, EventWithOriginatorID):
            pass

        # Check construction requires both an ID and version.
        with self.assertRaises(TypeError):
            # noinspection PyArgumentList
            Event()

        with self.assertRaises(TypeError):
            # noinspection PyArgumentList
            Event(originator_id='1')

        with self.assertRaises(TypeError):
            Event(originator_version=1)

        event1 = Event(originator_id='1', originator_version=0)

        event2 = Event(originator_id='1', originator_version=1)

        # Check the event attributes.
        self.assertEqual(event1.originator_id, '1')
        self.assertEqual(event2.originator_id, '1')
        self.assertEqual(event1.originator_version, 0)
        self.assertEqual(event2.originator_version, 1)

        # Check the events are not equal to each other, whilst being equal to themselves.
        self.assertEqual(event1, event1)
        self.assertEqual(event2, event2)
        self.assertNotEqual(event1, event2)


class TestEventWithTimestampAndOriginatorID(unittest.TestCase):
    def test_one_subclass(self):
        # Check base class can be sub-classed.
        class Event(EventWithTimestamp, EventWithOriginatorID):
            pass

        # Check construction requires an ID.
        with self.assertRaises(TypeError):
            Event()

        # Get timestamp before events.
        time1 = decimaltimestamp()

        # Construct events.
        event1 = Event(originator_id='1')
        event2 = Event(originator_id='1')

        # Check the entity IDs.
        self.assertEqual(event1.originator_id, '1')
        self.assertEqual(event2.originator_id, '1')

        # Check the event timestamps.
        self.assertLess(time1, event1.timestamp)
        self.assertLess(event1.timestamp, event2.timestamp)
        self.assertLess(event2.timestamp, decimaltimestamp())

        # Check the events are not equal to each other, whilst being equal to themselves.
        self.assertEqual(event1, event1)
        self.assertEqual(event2, event2)
        self.assertNotEqual(event1, event2)

    def test_two_subclasses(self):
        # Check base class can be sub-classed.
        class Event(EventWithTimestamp, EventWithOriginatorID):
            pass

        class Event2(EventWithTimestamp, EventWithOriginatorID):
            pass

        # Check subclass can be instantiated with 'originator_id' parameter.
        ID1 = uuid4()
        ID2 = uuid4()
        VALUE1 = 'a string'
        VALUE2 = 'another string'
        event1 = Event(
            originator_id=ID1,
        )
        self.assertEqual(event1.originator_id, ID1)

        # Check event has a domain event ID, and a timestamp.
        self.assertTrue(event1.timestamp)
        self.assertIsInstance(event1.timestamp, Decimal)

        # Check subclass can be instantiated with 'timestamp' parameter.
        DOMAIN_EVENT_ID1 = create_timesequenced_event_id()
        event2 = Event(
            originator_id=ID1,
            timestamp=DOMAIN_EVENT_ID1,
        )
        self.assertEqual(event2.timestamp, DOMAIN_EVENT_ID1)

        # Check subclass can be instantiated with other parameters.
        event3 = Event(
            originator_id=ID1,
            an_attribute=VALUE1,
        )
        self.assertEqual(event3.an_attribute, VALUE1)

        # Check the attribute value cannot be changed.
        with self.assertRaises(AttributeError):
            event3.an_attribute = VALUE2
        self.assertEqual(event3.an_attribute, VALUE1)

        # Check an event is equal to itself.
        self.assertTrue(event3 == event3)
        self.assertFalse(event3 != event3)
        event4 = Event(
            originator_id=ID1,
            timestamp=event3.timestamp,
            an_attribute=VALUE1,
        )
        self.assertEqual(event3, event4)

        # Check domain events with same domain event ID have the same timestamp.
        event5 = Event(
            originator_id=event1.originator_id,
            timestamp=event1.timestamp,
            an_attribute=VALUE1,
        )
        self.assertEqual(event1.timestamp, event5.timestamp)

        # Check it's not equal to different type with same values.
        self.assertFalse(event3 == Event2(
            originator_id=event3.originator_id,
            an_attribute=event3.an_attribute,
        ))
        self.assertTrue(event3 != Event2(
            originator_id=ID1,
            an_attribute=VALUE1,
        ))

        # Check it's not equal to same type with different values.
        # - different originator_id
        self.assertNotEqual(event3.originator_id, ID2)
        self.assertEqual(event3.an_attribute, VALUE1)
        self.assertFalse(event2 == Event(
            originator_id=ID2,
            an_attribute=VALUE1,
        ))
        # - different attribute value
        self.assertTrue(event2 != Event(
            originator_id=ID1,
            an_attribute=VALUE2,
        ))

        # Check domain events with different domain event IDs have different timestamps.
        event4 = Event(
            originator_id=ID1,
            an_attribute=VALUE1,
        )
        self.assertNotEqual(event2.timestamp, event4.timestamp)
        self.assertNotEqual(event2.timestamp, event4.timestamp)


# Todo: Review and reduce. This is the original test case, much but not all of which is covered by the new tests above.
class TestEvents(unittest.TestCase):
    def tearDown(self):
        _event_handlers.clear()

    def test_event_attributes(self):
        entity_id1 = uuid4()
        event = Example.Created(
            originator_id=entity_id1,
            originator_topic=get_topic(Example),
            a=1, b=2
        )

        # Check constructor keyword args lead to read-only attributes.
        self.assertEqual(1, event.a)
        self.assertEqual(2, event.b)
        self.assertRaises(AttributeError, getattr, event, 'c')
        self.assertRaises(AttributeError, setattr, event, 'c', 3)

        # Check domain event has auto-generated timestamp.
        self.assertIsInstance(event.timestamp, Decimal)

        # Check timestamp value can be given to domain events.
        event1 = Example.Created(
            originator_id=entity_id1,
            originator_topic=get_topic(Example),
            a=1, b=2, timestamp=3
        )
        self.assertEqual(3, event1.timestamp)

    def test_publish_subscribe_unsubscribe(self):
        # Check subscribing event handlers with predicates.
        # - when predicate is True, handler should be called
        event = mock.Mock()
        predicate = mock.Mock()
        handler = mock.Mock()

        # Check we can assert there are no event handlers subscribed.
        assert_event_handlers_empty()

        # When predicate is True, handler should be called ONCE.
        subscribe(handler=handler, predicate=predicate)

        # Check we can assert there are event handlers subscribed.
        self.assertRaises(EventHandlersNotEmptyError, assert_event_handlers_empty)

        # Check what happens when an event is published.
        publish(event)
        predicate.assert_called_once_with(event)
        handler.assert_called_once_with(event)

        # When predicate is True, after unsubscribing, handler should NOT be called again.
        unsubscribe(handler=handler, predicate=predicate)
        publish(event)
        predicate.assert_called_once_with(event)
        handler.assert_called_once_with(event)

        # Check we can assert there are no event handlers subscribed.
        assert_event_handlers_empty()

        # When predicate is False, handler should NOT be called.
        predicate = lambda x: False
        handler = mock.Mock()
        subscribe(handler=handler, predicate=predicate)
        publish(event)
        self.assertEqual(0, handler.call_count)

        # Unsubscribe.
        unsubscribe(handler=handler, predicate=predicate)

        # Check we can assert there are no event handlers subscribed.
        assert_event_handlers_empty()

    def test_hash(self):
        entity_id1 = uuid4()
        event1 = Example.Created(
            originator_id=entity_id1,
            originator_topic=get_topic(Example),
            a=1, b=2, timestamp=3
        )
        event2 = Example.Created(
            originator_id=entity_id1,
            originator_topic=get_topic(Example),
            a=1, b=2, timestamp=3
        )
        self.assertEqual(hash(event1), hash(event2))

    def test_equality_comparison(self):
        entity_id1 = uuid4()
        event1 = Example.Created(
            originator_id=entity_id1,
            originator_topic=get_topic(Example),
            a=1, b=2, timestamp=3
        )
        event2 = Example.Created(
            originator_id=entity_id1,
            originator_topic=get_topic(Example),
            a=1, b=2, timestamp=3
        )
        event3 = Example.Created(
            originator_id=entity_id1,
            originator_topic=get_topic(Example),
            a=3, b=2, timestamp=3
        )
        self.assertEqual(event1, event2)
        self.assertNotEqual(event1, event3)
        self.assertNotEqual(event2, event3)
        self.assertNotEqual(event2, None)

    def test_repr(self):
        entity_id1 = uuid4()
        event1 = Example.Created(
            originator_id=entity_id1,
            originator_topic=get_topic(Example),
            a=1, b=2, timestamp=3
        )
        self.maxDiff = None
        self.assertEqual(
            ("Example.Created(__event_hash__='{}', "
             "__event_topic__='eventsourcing.example.domainmodel#Example.Created', a=1, b=2, "
             "originator_id={}, "
             "originator_topic='eventsourcing.example.domainmodel#Example', originator_version=0, timestamp=3)"
            ).format(event1.__event_hash__, repr(entity_id1)),
            repr(event1)
        )

    def test_topic_resolution_error(self):
        # Check topic resolution error is raised, if the module path is
        # broken, and if the class name is broken.
        resolve_topic('eventsourcing.domain.model.events#DomainEvent')
        with self.assertRaises(TopicResolutionError):
            resolve_topic('eventsourcing.domain.model.broken#DomainEvent')
        with self.assertRaises(TopicResolutionError):
            resolve_topic('eventsourcing.domain.model.events#Broken')



import multiprocessing
from multiprocessing import Manager
from time import sleep

import six

from eventsourcing.application.process import Prompt
from eventsourcing.application.system import System
from eventsourcing.domain.model.decorators import retry
from eventsourcing.domain.model.events import subscribe, unsubscribe
from eventsourcing.exceptions import CausalDependencyFailed
from eventsourcing.interface.notificationlog import RecordManagerNotificationLog

DEFAULT_POLL_INTERVAL = 5


class Multiprocess(object):

    def __init__(self, system, pipeline_ids=(-1,), poll_interval=None, notification_log_section_size=5,
                 pool_size=1, setup_tables=False, sleep_for_setup_tables=0):
        self.pool_size = pool_size
        self.system = system
        self.pipeline_ids = pipeline_ids
        self.poll_interval = poll_interval or DEFAULT_POLL_INTERVAL
        assert isinstance(system, System)
        self.os_processes = None
        self.notification_log_section_size = notification_log_section_size
        self.setup_tables = setup_tables or system.setup_tables
        self.sleep_for_setup_tables = sleep_for_setup_tables

    def start(self):
        assert self.os_processes is None, "Already started"

        self.os_processes = []

        self.manager = Manager()
        self.inboxes = {}
        self.outboxes = {}

        # Setup queues.
        for pipeline_id in self.pipeline_ids:
            for process_class, upstream_classes in self.system.followings.items():
                inbox_id = (pipeline_id, process_class.__name__.lower())
                if inbox_id not in self.inboxes:
                    self.inboxes[inbox_id] = self.manager.Queue()
                for upstream_class in upstream_classes:
                    outbox_id = (pipeline_id, upstream_class.__name__.lower())
                    if outbox_id not in self.outboxes:
                        self.outboxes[outbox_id] = Outbox()
                    if inbox_id not in self.outboxes[outbox_id].downstream_inboxes:
                        self.outboxes[outbox_id].downstream_inboxes[inbox_id] = self.inboxes[inbox_id]

        # Subscribe to broadcast prompts published by a process
        # application in the parent operating system process.
        subscribe(handler=self.broadcast_prompt, predicate=self.is_prompt)

        # Start operating system process.
        for pipeline_id in self.pipeline_ids:
            for process_class, upstream_classes in self.system.followings.items():
                os_process = OperatingSystemProcess(
                    application_process_class=process_class,
                    upstream_names=[cls.__name__.lower() for cls in upstream_classes],
                    poll_interval=self.poll_interval,
                    pipeline_id=pipeline_id,
                    notification_log_section_size=self.notification_log_section_size,
                    pool_size=self.pool_size,
                    setup_tables=self.setup_tables,
                    inbox=self.inboxes[(pipeline_id, process_class.__name__.lower())],
                    outbox=self.outboxes[(pipeline_id, process_class.__name__.lower())],
                )
                os_process.daemon = True
                os_process.start()
                self.os_processes.append(os_process)
                if self.setup_tables:
                    # Avoid conflicts when creating tables.
                    sleep(self.sleep_for_setup_tables)

    def broadcast_prompt(self, prompt):
        outbox_id = (prompt.pipeline_id, prompt.process_name)
        assert outbox_id in self.outboxes, (outbox_id, self.outboxes.keys())
        self.outboxes[outbox_id].put(prompt)

    @staticmethod
    def is_prompt(event):
        return isinstance(event, Prompt)

    def close(self):
        unsubscribe(handler=self.broadcast_prompt, predicate=self.is_prompt)

        for os_process in self.os_processes:
            os_process.inbox.put('QUIT')

        for os_process in self.os_processes:
            os_process.join(timeout=10)

        for os_process in self.os_processes:
            os_process.is_alive() and os_process.terminate()

        self.os_processes = None
        self.manager = None

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class Outbox(object):
    def __init__(self):
        self.downstream_inboxes = {}

    def put(self, msg):
        for q in self.downstream_inboxes.values():
            q.put(msg)


class OperatingSystemProcess(multiprocessing.Process):

    def __init__(self, application_process_class, upstream_names, pipeline_id=-1,
                 poll_interval=DEFAULT_POLL_INTERVAL, notification_log_section_size=None,
                 pool_size=5, setup_tables=False, inbox=None, outbox=None, *args, **kwargs):
        super(OperatingSystemProcess, self).__init__(*args, **kwargs)
        self.application_process_class = application_process_class
        self.upstream_names = upstream_names
        self.daemon = True
        self.pipeline_id = pipeline_id
        self.poll_interval = poll_interval
        self.notification_log_section_size = notification_log_section_size
        self.pool_size = pool_size
        self.inbox = inbox
        self.outbox = outbox
        self.setup_tables = setup_tables

    def run(self):
        # Construct process application object.
        self.process = self.application_process_class(
            pipeline_id=self.pipeline_id,
            notification_log_section_size=self.notification_log_section_size,
            pool_size=self.pool_size,
            setup_table=self.setup_tables,
        )

        # Follow upstream notification logs.
        for upstream_name in self.upstream_names:

            # Obtain a notification log object (local or remote) for the upstream process.
            if upstream_name == self.process.name:
                # Upstream is this process's application,
                # so use own notification log.
                notification_log = self.process.notification_log
            else:
                # For a different application, we need to construct a notification
                # log with a record manager that has the upstream application ID.
                # Currently assumes all applications are using the same database
                # and record manager class. If it wasn't the same database,we would
                # to use a remote notification log, and upstream would need to provide
                # an API from which we can pull. It's not unreasonable to have a fixed
                # number of application processes connecting to the same database.
                record_manager = self.process.event_store.record_manager
                notification_log = RecordManagerNotificationLog(
                    record_manager=record_manager.clone(
                        application_name=upstream_name,
                        pipeline_id=self.pipeline_id
                    ),
                    section_size=self.process.notification_log_section_size
                )
                # Todo: Support upstream partition IDs different from self.pipeline_id.
                # Todo: Support combining partitions. Read from different partitions but write to the same partition,
                # could be one os process that reads from many logs of the same upstream app, or many processes each
                # reading one partition with contention writing to the same partition).
                # Todo: Support dividing partitions Read from one but write to many. Maybe one process per
                # upstream partition, round-robin to pick partition for write. Or have many processes reading
                # with each taking it in turn to skip processing somehow.
                # Todo: Dividing partitions would allow a stream to flow at the same rate through slower
                # process applications.
                # Todo: Support merging results from "replicated state machines" - could have a command
                # logging process that takes client commands and presents them in a notification log.
                # Then the system could be deployed in different places, running independently, receiving
                # the same commands, and running the same processes. The command logging process could
                # be accompanied with a result logging process that reads results from replicas as they
                # are available. Not sure what to do if replicas return different things. If one replica
                # goes down, then it could resume by pulling events from another? Not sure what to do.
                # External systems could be modelled as commands.

            # Make the process follow the upstream notification log.
            self.process.follow(upstream_name, notification_log)

        # Subscribe to broadcast prompts published by the process application.
        subscribe(handler=self.broadcast_prompt, predicate=self.is_prompt)

        try:
            self.loop_on_prompts()
        finally:
            unsubscribe(handler=self.broadcast_prompt, predicate=self.is_prompt)

    @retry(CausalDependencyFailed, max_attempts=100, wait=0.1)
    def loop_on_prompts(self):

        # Run once, in case prompts were missed.
        self.process.run()

        # Loop on getting prompts.
        while True:
            try:
                # Todo: Make the poll interval gradually increase if there are only timeouts?
                item = self.inbox.get(timeout=self.poll_interval)
                self.inbox.task_done()

                if item == 'QUIT':
                    self.process.close()
                    break

                else:
                    self.process.run(item)

            except six.moves.queue.Empty:
                # Basically, we're polling after a timeout.
                self.process.run()

    def broadcast_prompt(self, prompt):
        self.outbox.put(prompt)

    @staticmethod
    def is_prompt(event):
        return isinstance(event, Prompt)



from eventsourcing.tests.example_application_tests.base import ExampleApplicationTestCase
from eventsourcing.tests.sequenced_item_tests.test_sqlalchemy_record_manager import \
    WithSQLAlchemyRecordManagers


class TestExampleApplicationWithSQLAlchemy(WithSQLAlchemyRecordManagers, ExampleApplicationTestCase):
    pass



import datetime
from decimal import Decimal
from json import JSONDecoder, JSONEncoder, dumps, loads
from uuid import UUID

import dateutil.parser
import six

if not six.PY2:
    from json import JSONDecodeError

from eventsourcing.utils.topic import get_topic, resolve_topic
from collections import deque


class ObjectJSONEncoder(JSONEncoder):

    def __init__(self, sort_keys=True, *args, **kwargs):
        super(ObjectJSONEncoder, self).__init__(sort_keys=sort_keys, *args, **kwargs)

    def iterencode(self, o, _one_shot=False):
        try:
            return super(ObjectJSONEncoder, self).iterencode(o, _one_shot=_one_shot)
        except:
            raise

    def default(self, obj):
        if isinstance(obj, UUID):
            return {'UUID': obj.hex}
        elif isinstance(obj, datetime.datetime):
            return {'ISO8601_datetime': obj.strftime('%Y-%m-%dT%H:%M:%S.%f%z')}
        elif isinstance(obj, datetime.date):
            return {'ISO8601_date': obj.isoformat()}
        elif isinstance(obj, datetime.time):
            return {'ISO8601_time': obj.strftime('%H:%M:%S.%f')}
        elif isinstance(obj, Decimal):
            return {
                '__decimal__': str(obj),
            }
        elif hasattr(obj, '__class__') and hasattr(obj, '__dict__'):
            topic = get_topic(obj.__class__)
            state = obj.__dict__.copy()
            return {
                '__class__': {
                    'topic': topic,
                    'state': state,
                }
            }
        elif hasattr(obj, '__class__') and hasattr(obj, '__slots__'):
            topic = get_topic(obj.__class__)
            state = {k:getattr(obj, k) for k in obj.__slots__}
            return {
                '__class__': {
                    'topic': topic,
                    'state': state,
                }
            }
        elif isinstance(obj, set):
            return {'__set__': sorted(list(obj))}
        elif isinstance(obj, deque):
            assert list(obj) == []
            return {'__deque__': []}

        # Let the base class default method raise the TypeError.
        return JSONEncoder.default(self, obj)


class ObjectJSONDecoder(JSONDecoder):
    def __init__(self, object_hook=None, **kwargs):
        super(ObjectJSONDecoder, self).__init__(object_hook=object_hook or self.from_jsonable, **kwargs)

    @classmethod
    def from_jsonable(cls, d):
        if 'ISO8601_datetime' in d:
            return cls._decode_datetime(d)
        elif 'ISO8601_date' in d:
            return cls._decode_date(d)
        elif 'UUID' in d:
            return cls._decode_uuid(d)
        elif '__decimal__' in d:
            return cls._decode_decimal(d)
        elif 'ISO8601_time' in d:
            return cls._decode_time(d)
        elif '__class__' in d:
            return cls._decode_object(d)
        elif '__deque__' in d:
            return deque([])
        elif '__set__' in d:
            return set(d['__set__'])
        return d

    @classmethod
    def _decode_time(cls, d):
        hour, minute, seconds = d['ISO8601_time'].split(':')
        second, microsecond = seconds.split('.')
        return datetime.time(int(hour), int(minute), int(second), int(microsecond))

    @classmethod
    def _decode_decimal(cls, d):
        return Decimal(d['__decimal__'])

    @staticmethod
    def _decode_date(d):
        return datetime.datetime.strptime(d['ISO8601_date'], '%Y-%m-%d').date()

    @staticmethod
    def _decode_datetime(d):
        return dateutil.parser.parse(d['ISO8601_datetime'])

    @staticmethod
    def _decode_uuid(d):
        return UUID(d['UUID'])

    @staticmethod
    def _decode_object(d):
        topic = d['__class__']['topic']
        state = d['__class__']['state']
        obj_class = resolve_topic(topic)
        obj = object.__new__(obj_class)
        if hasattr(obj, '__dict__'):
            obj.__dict__.update(state)
        else:
            for k,v in state.items():
                setattr(obj, k, v)
        return obj


def json_dumps(obj, cls=None):
    return dumps(
        obj,
        separators=(',', ':'),
        sort_keys=True,
        cls=cls,
    )


def json_loads(s, cls=None):
    # Python 3 introduces JSONDecodeError.
    if six.PY2:
        return loads(s, cls=cls)
    else:
        try:
            return loads(s, cls=cls)
        except JSONDecodeError:
            raise ValueError("Couldn't load JSON string: {}".format(s))



from eventsourcing.domain.model.entity import AbstractEntityRepository
from eventsourcing.exceptions import RepositoryKeyError
from eventsourcing.infrastructure.eventplayer import EventPlayer
from eventsourcing.infrastructure.snapshotting import entity_from_snapshot


class EventSourcedRepository(EventPlayer, AbstractEntityRepository):
    def __contains__(self, entity_id):
        """
        Returns a boolean value according to whether entity with given ID exists.
        """
        return self.get_entity(entity_id) is not None

    def __getitem__(self, entity_id):
        """
        Returns entity with given ID.
        """
        # # Get entity from the cache.
        # if self._use_cache:
        #     try:
        #         return self._cache[entity_id]
        #     except KeyError:
        #         pass

        # Reconstitute the entity.
        entity = self.get_entity(entity_id)

        # Never created or already discarded?
        if entity is None:
            raise RepositoryKeyError(entity_id)

        # # Put entity in the cache.
        # if self._use_cache:
        #     self._cache[entity_id] = entity

        # Return entity.
        return entity

    def get_entity(self, entity_id, at=None):
        """
        Returns entity with given ID, optionally until position.
        """

        # Get a snapshot (None if none exist).
        if self._snapshot_strategy is not None:
            snapshot = self._snapshot_strategy.get_snapshot(entity_id, lte=at)
        else:
            snapshot = None

        # Decide the initial state of the entity, and the
        # version of the last item applied to the entity.
        if snapshot is None:
            initial_state = None
            gt = None
        else:
            initial_state = entity_from_snapshot(snapshot)
            gt = snapshot.originator_version

        # Obtain and return current state.
        return self.get_and_project_events(entity_id, gt=gt, lte=at, initial_state=initial_state)

    def get_and_project_events(self, entity_id, gt=None, gte=None, lt=None, lte=None, limit=None, initial_state=None,
                               query_descending=False):
        """
        Reconstitutes requested domain entity from domain events found in event store.
        """
        # Decide if query is in ascending order.
        #  - A "speed up" for when events are stored in descending order (e.g.
        #  in Cassandra) and it is faster to get them in that order.
        #  - This isn't useful when 'until' or 'after' or 'limit' are set,
        #    because the inclusiveness or exclusiveness of until and after
        #    and the end of the stream that is truncated by limit both depend on
        #    the direction of the query. Also paging backwards isn't useful, because
        #    all the events are needed eventually, so it would probably slow things
        #    down. Paging is intended to support replaying longer event streams, and
        #    only makes sense to work in ascending order.
        if gt is None and gte is None and lt is None and lte is None and self.__page_size__ is None:
            is_ascending = False
        else:
            is_ascending = not query_descending

        # Get entity's domain events from the event store.
        domain_events = self.event_store.get_domain_events(
            originator_id=entity_id,
            gt=gt,
            gte=gte,
            lt=lt,
            lte=lte,
            limit=limit,
            is_ascending=is_ascending,
            page_size=self.__page_size__
        )

        # The events will be replayed in ascending order.
        if not is_ascending:
            domain_events = list(reversed(list(domain_events)))

        # Project the domain events onto the initial state.
        return self.project_events(initial_state, domain_events)

    # Todo: Does this method belong on this class?
    def take_snapshot(self, entity_id, lt=None, lte=None):
        """
        Takes a snapshot of the entity as it existed after the most recent
        event, optionally less than, or less than or equal to, a particular position.
        """
        snapshot = None
        if self._snapshot_strategy:
            # Get the latest event (optionally until a particular position).
            latest_event = self.event_store.get_most_recent_event(entity_id, lt=lt, lte=lte)

            # If there is something to snapshot, then look for a snapshot
            # taken before or at the entity version of the latest event. Please
            # note, the snapshot might have a smaller version number than
            # the latest event if events occurred since the latest snapshot was taken.
            if latest_event is not None:
                latest_snapshot = self._snapshot_strategy.get_snapshot(
                    entity_id, lt=lt, lte=lte
                )
                latest_version = latest_event.originator_version

                if latest_snapshot and latest_snapshot.originator_version == latest_version:
                    # If up-to-date snapshot exists, there's nothing to do.
                    snapshot = latest_snapshot
                else:
                    # Otherwise recover entity state from latest snapshot.
                    if latest_snapshot:
                        initial_state = entity_from_snapshot(latest_snapshot)
                        gt = latest_snapshot.originator_version
                    else:
                        initial_state = None
                        gt = None

                    # Fast-forward entity state to latest version.
                    entity = self.get_and_project_events(
                        entity_id=entity_id,
                        gt=gt,
                        lte=latest_version,
                        initial_state=initial_state,
                    )

                    # Take snapshot from entity.
                    snapshot = self._snapshot_strategy.take_snapshot(entity_id, entity, latest_version)

        return snapshot



from collections import OrderedDict


class System(object):
    def __init__(self, *pipelines, **kwargs):
        """
        Initialises a "process network" system object.

        :param pipelines: Pipelines of process classes.

        Each pipeline of process classes shows directly which process
        follows which other process in the system.

        For example, the pipeline (A | B | C) shows that B follows A,
        and C follows B.

        The pipeline (A | A) shows that A follows A.

        The pipeline (A | B | A) shows that B follows A, and A follows B.

        The pipelines ((A | B | A), (A | C | A)) is equivalent to (A | B | A | C | A).
        """
        self.pipelines = pipelines
        self.process_classes = []
        for pipeline in self.pipelines:
            for process_class in pipeline:
                if process_class not in self.process_classes:
                    self.process_classes.append(process_class)
        self.processes_by_name = None
        self.is_session_shared = True

        # Determine which process follows which.
        self.followings = OrderedDict()
        for pipeline in self.pipelines:
            previous_class = None
            for process_class in pipeline:
                # Follower follows the followed.
                try:
                    follows = self.followings[process_class]
                except KeyError:
                    follows = []
                    self.followings[process_class] = follows

                if previous_class is not None and previous_class not in follows:
                    follows.append(previous_class)

                previous_class = process_class

        self.setup_tables = kwargs.get('setup_tables', False)

    # Todo: Extract function to new 'SinglethreadRunner' class or something (like the 'Multiprocess' class)?
    def setup(self):
        assert self.processes_by_name is None, "Already running"
        self.processes_by_name = {}

        # Construct the processes.
        session = None
        for process_class in self.process_classes:
            kwargs = {}
            kwargs['setup_table'] = self.setup_tables
            if process_class.is_constructed_with_session:
                kwargs['session'] = session
            process = process_class(**kwargs)
            self.processes_by_name[process.name] = process
            if process_class.is_constructed_with_session and self.is_session_shared:
                if session is None:
                    session = process.session

        # Configure which process follows which.
        for follower_class, follows in self.followings.items():
            follower = self.processes_by_name[follower_class.__name__.lower()]
            for followed_class in follows:
                followed = self.processes_by_name[followed_class.__name__.lower()]
                follower.follow(followed.name, followed.notification_log)

    def __getattr__(self, process_name):
        assert self.processes_by_name is not None, "Not running"
        return self.processes_by_name[process_name]

    def close(self):
        assert self.processes_by_name is not None, "Not running"
        for process in self.processes_by_name.values():
            process.close()
        self.processes_by_name = None

    def __enter__(self):
        self.setup()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



from eventsourcing.application.command import CommandProcess
from eventsourcing.application.process import ProcessApplication
from eventsourcing.application.simple import SimpleApplication

from eventsourcing.infrastructure.django.manager import DjangoRecordManager
from eventsourcing.infrastructure.django.models import StoredEventRecord


class ApplicationWithDjango(SimpleApplication):
    record_manager_class = DjangoRecordManager
    stored_event_record_class = StoredEventRecord


class ProcessApplicationWithDjango(ApplicationWithDjango, ProcessApplication):
    pass


class CommandProcessWithDjango(ApplicationWithDjango, CommandProcess):
    pass


class SimpleApplication(ApplicationWithDjango):
    """Shorter name for ApplicationWithDjango."""


class ProcessApplication(ProcessApplicationWithDjango):
    """Shorter name for ProcessApplicationWithDjango."""


class CommandProcess(CommandProcessWithDjango):
    """Shorter name for CommandProcessWithDjango."""



import base64
import zlib

from Crypto.Cipher import AES

from eventsourcing.exceptions import DataIntegrityError
from eventsourcing.utils.random import random_bytes


class AESCipher(object):
    """
    Cipher strategy that uses Crypto library AES cipher in GCM mode.
    """

    def __init__(self, cipher_key):
        """
        Initialises AES cipher strategy with ``cipher_key``.

        :param cipher_key: 16, 24, or 32 random bytes
        """
        assert len(cipher_key) in [16, 24, 32]
        self.cipher_key = cipher_key

    def encrypt(self, plaintext):
        """Return ciphertext for given plaintext."""

        # String to bytes.
        plainbytes = plaintext.encode('utf8')

        # Compress plaintext bytes.
        compressed = zlib.compress(plainbytes)

        # Construct AES-GCM cipher, with 96-bit nonce.
        cipher = AES.new(self.cipher_key, AES.MODE_GCM, nonce=random_bytes(12))

        # Encrypt and digest.
        encrypted, tag = cipher.encrypt_and_digest(compressed)

        # Combine with nonce.
        combined = cipher.nonce + tag + encrypted

        # Encode as Base64.
        cipherbytes = base64.b64encode(combined)

        # Bytes to string.
        ciphertext = cipherbytes.decode('utf8')

        # Return ciphertext.
        return ciphertext

    def decrypt(self, ciphertext):
        """Return plaintext for given ciphertext."""

        # String to bytes.
        cipherbytes = ciphertext.encode('utf8')

        # Decode from Base64.
        try:
            combined = base64.b64decode(cipherbytes)
        except (base64.binascii.Error, TypeError) as e:
            # base64.binascii.Error for Python 3.
            # TypeError for Python 2.
            raise DataIntegrityError("Cipher text is damaged: {}".format(e))

        # Split out the nonce, tag, and encrypted data.
        nonce = combined[:12]
        if len(nonce) != 12:
            raise DataIntegrityError("Cipher text is damaged: invalid nonce length")

        tag = combined[12:28]
        if len(tag) != 16:
            raise DataIntegrityError("Cipher text is damaged: invalid tag length")

        encrypted = combined[28:]

        # Construct AES cipher, with old nonce.
        cipher = AES.new(self.cipher_key, AES.MODE_GCM, nonce)

        # Decrypt and verify.
        try:
            compressed = cipher.decrypt_and_verify(encrypted, tag)
        except ValueError as e:
            raise DataIntegrityError("Cipher text is damaged: {}".format(e))

        # Decompress plaintext bytes.
        plainbytes = zlib.decompress(compressed)

        # Bytes to string.
        plaintext = plainbytes.decode('utf8')

        # Return plaintext.
        return plaintext



from eventsourcing.infrastructure.cassandra.manager import CassandraRecordManager
from eventsourcing.infrastructure.cassandra.records import IntegerSequencedRecord, SnapshotRecord, \
    TimestampSequencedRecord
from eventsourcing.infrastructure.factory import InfrastructureFactory


class CassandraInfrastructureFactory(InfrastructureFactory):
    record_manager_class = CassandraRecordManager
    integer_sequenced_record_class = IntegerSequencedRecord
    timestamp_sequenced_record_class = TimestampSequencedRecord
    snapshot_record_class = SnapshotRecord



from sqlalchemy import asc, bindparam, desc, select, text
from sqlalchemy.exc import DBAPIError, IntegrityError
from sqlalchemy.orm.exc import MultipleResultsFound, NoResultFound
from sqlalchemy.sql import func

from eventsourcing.exceptions import ProgrammingError
from eventsourcing.infrastructure.base import AbstractTrackingRecordManager, RelationalRecordManager
from eventsourcing.infrastructure.sqlalchemy.records import NotificationTrackingRecord


class SQLAlchemyRecordManager(RelationalRecordManager):
    tracking_record_class = NotificationTrackingRecord

    def __init__(self, session, *args, **kwargs):
        super(SQLAlchemyRecordManager, self).__init__(*args, **kwargs)
        self.session = session

    def _prepare_insert(self, tmpl, record_class, field_names, placeholder_for_id=False):
        """
        With transaction isolation level of "read committed" this should
        generate records with a contiguous sequence of integer IDs, assumes
        an indexed ID column, the database-side SQL max function, the
        insert-select-from form, and optimistic concurrency control.
        """
        field_names = list(field_names)
        if hasattr(record_class, 'application_name') and 'application_name' not in field_names:
            field_names.append('application_name')
        if hasattr(record_class, 'pipeline_id') and 'pipeline_id' not in field_names:
            field_names.append('pipeline_id')
        if hasattr(record_class, 'causal_dependencies') and 'causal_dependencies' not in field_names:
            field_names.append('causal_dependencies')
        if hasattr(record_class, 'id') and placeholder_for_id and 'id' not in field_names:
            field_names.append('id')

        statement = text(tmpl.format(
            tablename=self.get_record_table_name(record_class),
            columns=", ".join(field_names),
            placeholders=", ".join([":{}".format(f) for f in field_names]),
        ))

        # Define bind parameters with explicit types taken from record column types.
        bindparams = []
        for col_name in field_names:
            column_type = getattr(record_class, col_name).type
            bindparams.append(bindparam(col_name, type_=column_type))

        # Redefine statement with explicitly typed bind parameters.
        statement = statement.bindparams(*bindparams)

        # Compile the statement with the session dialect.
        compiled = statement.compile(dialect=self.session.bind.dialect)

        return compiled

    def write_records(self, records, tracking_kwargs=None):
        try:
            with self.session.bind.begin() as connection:
                if tracking_kwargs:
                    # Add tracking record to session.
                    params = {c: tracking_kwargs[c] for c in self.tracking_record_field_names}
                    connection.execute(self.insert_tracking_record, **params)

                for record in records:

                    params = {c: getattr(record, c) for c in self.field_names}
                    if hasattr(self.record_class, 'application_name'):
                        params['application_name'] = self.application_name
                    if hasattr(self.record_class, 'pipeline_id'):
                        params['pipeline_id'] = self.pipeline_id
                    if hasattr(record, 'causal_dependencies'):
                        params['causal_dependencies'] = record.causal_dependencies

                    statement = self.insert_values

                    if hasattr(self.record_class, 'id'):
                        if record.id is None and self.contiguous_record_ids:
                            # Do an "insert select max" from existing.
                            statement = self.insert_select_max
                        elif record.id == '':
                            # Don't want to put this in the notification log.
                            params['id'] = None
                            if hasattr(self.record_class, 'pipeline_id'):
                                params['pipeline_id'] = None
                        else:
                            # ID can't be auto-incrementing if table has an application_name.
                            if hasattr(self.record_class, 'application_name'):
                                # We need a value and don't have one.
                                assert record.id, "record ID not set when required"
                            params['id'] = record.id

                    connection.execute(statement, **params)

        except IntegrityError as e:
            self.raise_record_integrity_error(e)

        except DBAPIError as e:
            self.raise_operational_error(e)

    def get_records(self, sequence_id, gt=None, gte=None, lt=None, lte=None, limit=None,
                    query_ascending=True, results_ascending=True):
        assert limit is None or limit >= 1, limit
        try:
            # Filter by sequence_id.
            filter_kwargs = {
                self.field_names.sequence_id: sequence_id
            }
            # Optionally, filter by application_name.
            if hasattr(self.record_class, 'application_name'):
                filter_kwargs['application_name'] = self.application_name
            query = self.filter_by(**filter_kwargs)

            # Filter and order by position.
            position_field = getattr(self.record_class, self.field_names.position)

            if query_ascending:
                query = query.order_by(asc(position_field))
            else:
                query = query.order_by(desc(position_field))

            if gt is not None:
                query = query.filter(position_field > gt)
            if gte is not None:
                query = query.filter(position_field >= gte)
            if lt is not None:
                query = query.filter(position_field < lt)
            if lte is not None:
                query = query.filter(position_field <= lte)

            # Limit the results.
            if limit is not None:
                query = query.limit(limit)

            # Get the result set.
            results = query.all()

        finally:
            self.session.close()

        # Reverse if necessary.
        if results_ascending != query_ascending:
            # This code path is under test, but not otherwise used ATM.
            results.reverse()

        return results

    def get_notifications(self, start=None, stop=None, *args, **kwargs):
        try:
            query = self.orm_query()
            if hasattr(self.record_class, 'application_name'):
                query = query.filter(self.record_class.application_name == self.application_name)
            if hasattr(self.record_class, 'pipeline_id'):
                query = query.filter(self.record_class.pipeline_id == self.pipeline_id)
            if hasattr(self.record_class, 'id'):
                query = query.order_by(asc('id'))
                # NB '+1' because record IDs start from 1.
                if start is not None:
                    query = query.filter(self.record_class.id >= start + 1)
                if stop is not None:
                    query = query.filter(self.record_class.id < stop + 1)
            # Todo: Should some tables with an ID not be ordered by ID?
            # Todo: Which order do other tables have?
            return query.all()
        finally:
            self.session.close()

    def get_item(self, sequence_id, position):
        return self.from_record(self.get_record(sequence_id, position))

    def get_pipeline_and_notification_id(self, sequence_id, position):
        # Todo: Optimise query by selecting only two columns: pipeline_id and id (notification ID).
        record = self.get_record(sequence_id, position)
        return record.pipeline_id, record.id

    def get_record(self, sequence_id, position):
        try:
            filter_args = {
                self.field_names.sequence_id: sequence_id
            }

            query = self.filter_by(**filter_args)
            if hasattr(self.record_class, 'application_name'):
                query = query.filter(self.record_class.application_name == self.application_name)

            position_field = getattr(self.record_class, self.field_names.position)

            query = query.filter(position_field == position)
            return query.one()
        except (NoResultFound, MultipleResultsFound):
            raise IndexError

    def filter_by(self, **kwargs):
        return self.orm_query().filter_by(**kwargs)

    def orm_query(self):
        return self.session.query(self.record_class)

    def get_max_record_id(self):
        try:
            query = self.session.query(func.max(self.record_class.id))
            if hasattr(self.record_class, 'application_name'):
                query = query.filter(self.record_class.application_name == self.application_name)
            if hasattr(self.record_class, 'pipeline_id'):
                query = query.filter(self.record_class.pipeline_id == self.pipeline_id)

            return query.scalar()
        finally:
            self.session.close()

    def all_sequence_ids(self):
        c = self.record_class.__table__.c
        sequence_id_col = getattr(c, self.field_names.sequence_id)
        expr = select([sequence_id_col], distinct=True)

        if hasattr(self.record_class, 'application_name'):
            expr = expr.where(c.application_name == self.application_name)

        try:
            for row in self.session.query(expr):
                yield row[0]
        finally:
            self.session.close()

    def delete_record(self, record):
        """
        Permanently removes record from table.
        """
        try:
            self.session.delete(record)
            self.session.commit()
        except Exception as e:
            self.session.rollback()
            raise ProgrammingError(e)
        finally:
            self.session.close()

    def get_record_table_name(self, record_class):
        return record_class.__table__.name

    def clone(self, **kwargs):
        return super(SQLAlchemyRecordManager, self).clone(session=self.session, **kwargs)


class TrackingRecordManager(AbstractTrackingRecordManager):
    record_class = NotificationTrackingRecord

    def __init__(self, session):
        self.session = session

    def get_max_record_id(self, application_name, upstream_application_name, pipeline_id):
        query = self.session.query(func.max(self.record_class.notification_id))
        query = query.filter(self.record_class.application_name == application_name)
        query = query.filter(self.record_class.upstream_application_name == upstream_application_name)
        query = query.filter(self.record_class.pipeline_id == pipeline_id)
        return query.scalar()

    def has_tracking_record(self, application_name, upstream_application_name, pipeline_id, notification_id):
        query = self.session.query(self.record_class)
        query = query.filter(self.record_class.application_name == application_name)
        query = query.filter(self.record_class.upstream_application_name == upstream_application_name)
        query = query.filter(self.record_class.pipeline_id == pipeline_id)
        query = query.filter(self.record_class.notification_id == notification_id)
        try:
            query.one()
        except (MultipleResultsFound, NoResultFound):
            return False
        else:
            return True



from eventsourcing.application.policies import SnapshottingPolicy
from eventsourcing.application.simple import SimpleApplication
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.snapshotting import EventSourcedSnapshotStrategy


class ApplicationWithSnapshotting(SimpleApplication):
    # Todo: Change this to default to None?
    snapshot_period = 2

    def __init__(self, snapshot_period=None, snapshot_record_class=None, **kwargs):
        self.snapshot_period = snapshot_period or self.snapshot_period
        self.snapshot_record_class = snapshot_record_class
        self.snapshotting_policy = None
        super(ApplicationWithSnapshotting, self).__init__(**kwargs)

    def setup_event_store(self):
        super(ApplicationWithSnapshotting, self).setup_event_store()
        # Setup event store for snapshots.
        self.snapshot_store = EventStore(
            record_manager=self.infrastructure_factory.construct_snapshot_record_manager(),
            sequenced_item_mapper=self.sequenced_item_mapper_class(
                sequenced_item_class=self.sequenced_item_class
            )
        )

    def setup_repository(self, **kwargs):
        # Setup repository with a snapshot strategy.
        self.snapshot_strategy = EventSourcedSnapshotStrategy(
            snapshot_store=self.snapshot_store
        )
        super(ApplicationWithSnapshotting, self).setup_repository(
            snapshot_strategy=self.snapshot_strategy, **kwargs
        )

    def setup_persistence_policy(self):
        super(ApplicationWithSnapshotting, self).setup_persistence_policy()
        self.snapshotting_policy = SnapshottingPolicy(
            repository=self.repository,
            snapshot_store=self.snapshot_store,
            persist_event_type=self.persist_event_type,
            period=self.snapshot_period
        )

    def setup_table(self):
        super(ApplicationWithSnapshotting, self).setup_table()
        if self.datastore is not None:
            self.datastore.setup_table(self.snapshot_store.record_manager.record_class)

    def close(self):
        super(ApplicationWithSnapshotting, self).close()
        if self.snapshotting_policy is not None:
            self.snapshotting_policy.close()
            self.snapshotting_policy = None



import itertools
from copy import deepcopy
from uuid import uuid4

import six
from eventsourcing.application.sqlalchemy import ProcessApplicationWithSnapshotting
from eventsourcing.application.system import System
from eventsourcing.contrib.paxos.composable import PaxosInstance, Resolution, PaxosMessage
from eventsourcing.domain.model.aggregate import AggregateRoot
from eventsourcing.domain.model.decorators import attribute


class PaxosAggregate(AggregateRoot):
    def __init__(self, quorum_size, network_uid, *args, **kwargs):
        assert isinstance(quorum_size, six.integer_types)
        self.quorum_size = quorum_size
        self.network_uid = network_uid
        self.promises_received = set()
        self.nacks_received = set()
        self.leader = False
        self.proposals = {}
        self.acceptors = {}
        # self.resolution = None
        super(AggregateRoot, self).__init__(*args, **kwargs)

    paxos_attrs = [
        'proposed_value',
        'proposal_id',
        'promised_id',
        'accepted_id',
        'accepted_value',
        'highest_proposal_id',
        'promises_received',
        'nacks_received',
        'highest_accepted_id',
        'leader',
        'proposals',
        'acceptors',
        'final_value',
        'final_acceptors',
        'final_proposal_id',
    ]

    @property
    def paxos(self):
        instance = PaxosInstance(
            network_uid=self.network_uid,
            quorum_size=self.quorum_size,
        )

        for attr in self.paxos_attrs:
            value = getattr(self, attr, None)
            if value is not None:
                if isinstance(value, (set, list, dict, tuple)):
                    value = deepcopy(value)
                setattr(instance, attr, value)

        return instance

    @attribute
    def resolution(self):
        pass

    class Event(AggregateRoot.Event):
        pass

    class Started(Event, AggregateRoot.Created):
        pass

    class AttributesChanged(Event):
        __notifiable__ = False

        def __init__(self, changes=None, **kwargs):
            super(PaxosAggregate.AttributesChanged, self).__init__(
                changes=changes, **kwargs
            )

        @property
        def changes(self):
            return self.__dict__['changes']

        def mutate(self, obj):
            for attr, value in self.changes.items():
                setattr(obj, attr, value)

    class MessageAnnounced(Event):
        @property
        def msg(self):
            return self.__dict__['msg']

    @classmethod
    def start(cls, originator_id, quorum_size, network_uid):
        assert isinstance(quorum_size, six.integer_types), "Not an integer: {}".format(quorum_size)
        return cls.__create__(
            event_class=cls.Started,
            originator_id=originator_id,
            quorum_size=quorum_size,
            network_uid=network_uid
        )

    def propose_value(self, value, assume_leader=False):
        paxos = self.paxos
        paxos.leader = assume_leader
        msg = paxos.propose_value(value)
        if msg is None:
            msg = paxos.prepare()
        self.setattrs_from_paxos(paxos)
        self.announce(msg)
        return msg

    def receive_message(self, msg):
        paxos = self.paxos
        while msg:

            if isinstance(msg, Resolution):
                if self.resolution is None:
                    print("{} resolved value {}".format(self.network_uid, msg.value))
                    self.resolution = msg
                break
            else:
                print("{} <- {} <- {}".format(self.network_uid, msg.__class__.__name__, msg.from_uid))
                msg = paxos.receive(msg)
                if msg:
                    if not isinstance(msg, Resolution):
                        self.announce(msg)

        self.setattrs_from_paxos(paxos)

    def announce(self, msg):
        print("{} -> {}".format(self.network_uid, msg.__class__.__name__))

        self.__trigger_event__(
            event_class=self.MessageAnnounced,
            msg=msg,
        )

    def setattrs_from_paxos(self, paxos):
        changes = {}
        for attr in self.paxos_attrs:
            paxos_value = getattr(paxos, attr)
            if paxos_value != getattr(self, attr, None):
                print("{} {}: {}".format(self.network_uid, attr, paxos_value))
                changes[attr] = paxos_value
                setattr(self, attr, paxos_value)
        if changes:
            self.__trigger_event__(
                event_class=self.AttributesChanged,
                changes=changes
            )


class PaxosProcess(ProcessApplicationWithSnapshotting):
    always_track_notifications = True
    snapshot_period = None
    persist_event_type = PaxosAggregate.Event

    def propose_value(self, value, quorum_size, assume_leader=False):
        paxos = PaxosAggregate.start(
            originator_id=uuid4(),
            quorum_size=quorum_size,
            network_uid=self.name
        )
        msg = paxos.propose_value(value, assume_leader=assume_leader)
        while msg:
            msg = paxos.receive_message(msg)
        self.record_new_events(paxos)
        self.publish_prompt()
        self.repository.take_snapshot(paxos.id)
        return paxos

    def policy(self, repository, event):
        # print("Running policy for {}".format(self.name))
        if isinstance(event, PaxosAggregate.Started):
            if event.originator_id not in repository:
                return PaxosAggregate.start(
                    originator_id=event.originator_id,
                    quorum_size=event.quorum_size,
                    network_uid=self.name
                )
        elif isinstance(event, PaxosAggregate.MessageAnnounced):
            msg = event.msg
            assert isinstance(msg, PaxosMessage)
            paxos = repository[event.originator_id]
            assert isinstance(paxos, PaxosAggregate)
            if not paxos.resolution:
                paxos.receive_message(msg)
                self.repository.take_snapshot(paxos.id)


class PaxosSystem(System):
    def __init__(self, processes=3, **kwargs):
        classes = [type('PaxosProcess{}'.format(i), (PaxosProcess,), {}) for i in range(processes)]
        if processes > 1:
            pipelines = [[c[0], c[1], c[0]] for c in itertools.combinations(classes, 2)]
        else:
            pipelines = [classes]
        self.quorum_size = (processes + 2) // 2
        super(PaxosSystem, self).__init__(*pipelines, **kwargs)



import datetime
import os
import unittest
from time import sleep

from eventsourcing.application.multiprocess import Multiprocess
from eventsourcing.contrib.paxos.application import PaxosProcess, PaxosSystem
from eventsourcing.domain.model.events import assert_event_handlers_empty
from eventsourcing.tests.test_system_fixtures import set_db_uri


class TestPaxosSystem(unittest.TestCase):

    def test_single_threaded(self):
        system = PaxosSystem(setup_tables=True, processes=3)

        with system:
            paxos1 = system.paxosprocess0.propose_value(11111, quorum_size=system.quorum_size, assume_leader=False)
            paxos2 = system.paxosprocess1.propose_value(22222, quorum_size=system.quorum_size, assume_leader=False)
            paxos3 = system.paxosprocess2.propose_value(33333, quorum_size=system.quorum_size, assume_leader=False)

            # Check each process has a resolution.
            self.assertEqual(system.paxosprocess0.repository[paxos1.id].resolution.value, 11111)
            self.assertEqual(system.paxosprocess1.repository[paxos1.id].resolution.value, 11111)
            self.assertEqual(system.paxosprocess2.repository[paxos1.id].resolution.value, 11111)
            self.assertEqual(system.paxosprocess0.repository[paxos2.id].resolution.value, 22222)
            self.assertEqual(system.paxosprocess1.repository[paxos2.id].resolution.value, 22222)
            self.assertEqual(system.paxosprocess2.repository[paxos2.id].resolution.value, 22222)
            self.assertEqual(system.paxosprocess0.repository[paxos3.id].resolution.value, 33333)
            self.assertEqual(system.paxosprocess1.repository[paxos3.id].resolution.value, 33333)
            self.assertEqual(system.paxosprocess2.repository[paxos3.id].resolution.value, 33333)

    def test_multiprocessing(self):
        set_db_uri()
        system = PaxosSystem(setup_tables=True, processes=3)

        with Multiprocess(system, pipeline_ids=[1, 2, 3]), PaxosProcess(name='paxosprocess0', pipeline_id=1) as \
            paxosprocess0:
            sleep(3)
            print("Running {}".format(datetime.datetime.now()))
            paxos1 = paxosprocess0.propose_value(11111, quorum_size=system.quorum_size, assume_leader=True)

            paxosprocess0.change_pipeline(2)
            paxos2 = paxosprocess0.propose_value(22222, quorum_size=system.quorum_size, assume_leader=True)

            paxosprocess0.change_pipeline(3)
            paxos3 = paxosprocess0.propose_value(33333, quorum_size=system.quorum_size, assume_leader=True)

            # Check each process has a resolution.
            while True:
                if paxosprocess0.repository[paxos1.id].resolution:
                    break
            self.assertEqual(paxosprocess0.repository[paxos1.id].resolution.value, 11111)
            print("")
            print("")
            print("")
            print("")
            print("Finished {}".format(datetime.datetime.now()))

            while True:
                if paxosprocess0.repository[paxos2.id].resolution:
                    break
            self.assertEqual(paxosprocess0.repository[paxos2.id].resolution.value, 22222)

            while True:
                if paxosprocess0.repository[paxos3.id].resolution:
                    break
            self.assertEqual(paxosprocess0.repository[paxos3.id].resolution.value, 33333)

    def tearDown(self):
        assert_event_handlers_empty()

        try:
            del (os.environ['DB_URI'])
        except KeyError:
            pass
